import uuid
from weaviate.classes.query import Filter, QueryReference

from elysia.objects import Retrieval

# Client Manager
from elysia.util.client import ClientManager

# Utilities
from elysia.util.parsing import format_dict_to_serialisable


class MessageRetrieval(Retrieval):
    def __init__(
        self,
        objects: list[dict],
        metadata: dict,
        mapping: dict | None = None,
        **kwargs,
    ) -> None:
        for obj in objects:
            obj["relevant"] = False

        Retrieval.__init__(
            self,
            objects,
            payload_type="message",
            metadata=metadata,
            mapping=mapping,
            unmapped_keys=[
                "uuid",
                "ELYSIA_SUMMARY",
                "collection_name",
                "relevant",
                "_REF_ID",
            ],
            **kwargs,
        )


class ConversationRetrieval(Retrieval):
    def __init__(
        self,
        objects: list[dict],
        metadata: dict,
        mapping: dict | None = None,
        **kwargs,
    ) -> None:
        Retrieval.__init__(
            self,
            objects,
            payload_type="conversation",
            metadata=metadata,
            mapping=mapping,
            unmapped_keys=[
                "uuid",
                "ELYSIA_SUMMARY",
                "collection_name",
                "relevant",
                "_REF_ID",
            ],
            **kwargs,
        )
        self.async_init_completed = False

    async def async_init(self, client_manager: ClientManager) -> None:
        if not self.async_init_completed:
            await self._return_all_messages_in_conversation(client_manager)

    async def _fetch_items_in_conversation(
        self,
        conversation_id: str,
        message_id: int,
        conversation_id_field_name: str,
        message_id_field_name: str,
        metadata: dict,
        client_manager: ClientManager,
    ) -> list[dict]:
        """
        Use Weaviate to fetch all messages in a conversation based on the conversation ID.
        """
        assert (
            "collection_name" in metadata
        ), "collection_name is required for fetching other messages in a conversation"

        async with client_manager.connect_to_async_client() as client:
            collection = client.collections.get(metadata["collection_name"])
            items_in_conversation = await collection.query.fetch_objects(
                filters=Filter.by_property(conversation_id_field_name).equal(
                    conversation_id
                )
            )

        output = []
        for obj in items_in_conversation.objects:
            output.append(obj.properties)
            output[-1]["uuid"] = str(obj.uuid)
            if output[-1][message_id_field_name] == message_id:
                output[-1]["relevant"] = True
            else:
                output[-1]["relevant"] = False

        return output

    async def _return_all_messages_in_conversation(
        self, client_manager: ClientManager
    ) -> list[dict]:
        """
        Return all messages in a conversation based on the response from Weaviate.
        """

        returned_objects = []
        conversations_seen = set()
        for o in self.objects:

            if o[self.mapping["conversation_id"]] in conversations_seen:
                continue

            conversations_seen.add(o[self.mapping["conversation_id"]])

            items_in_conversation = await self._fetch_items_in_conversation(
                o[self.mapping["conversation_id"]],
                o[self.mapping["message_id"]],
                self.mapping["conversation_id"],
                self.mapping["message_id"],
                self.metadata,
                client_manager,
            )
            # Check if all message_id values can be converted to int for sorting
            can_sort_as_int = True
            for x in items_in_conversation:
                try:
                    int(x[self.mapping["message_id"]])
                except (ValueError, TypeError, KeyError):
                    can_sort_as_int = False
                    break

            if can_sort_as_int:
                items_in_conversation.sort(
                    key=lambda x: int(x[self.mapping["message_id"]])
                )
            else:
                items_in_conversation.sort(key=lambda x: x[self.mapping["message_id"]])

            returned_objects.append(
                {
                    "messages": items_in_conversation,
                    "conversation_id": o[self.mapping["conversation_id"]],
                }
            )

        self.objects = returned_objects
        self.async_init_completed = True

        return returned_objects

    def to_json(self, mapping: bool = False) -> list[dict]:
        assert (
            self.async_init_completed
        ), "ConversationRetrieval not initialized, need to run .async_init(client_manager)"

        if self.mapping is not None and mapping:
            output_objects = []
            for obj in self.objects:  # outer level, conversation
                output_objects.append(
                    {
                        "conversation_id": obj["conversation_id"],
                        "messages": [],
                    }
                )
                for message in obj["messages"]:  # inner level, message
                    output_objects[-1]["messages"].append(
                        {
                            key: message[self.mapping[key]]
                            for key in self.mapping
                            if self.mapping[key] != ""
                        }
                    )
                    output_objects[-1]["messages"][-1].update(
                        {
                            key: message[key]
                            for key in self.unmapped_keys
                            if key in message
                        }
                    )
        else:
            output_objects = self.objects

        for object in output_objects:
            for message in object["messages"]:
                format_dict_to_serialisable(message)

        return output_objects


class DocumentRetrieval(Retrieval):
    def __init__(
        self,
        objects: list[dict],
        metadata: dict,
        mapping: dict | None = None,
        **kwargs,
    ) -> None:
        assert (
            "collection_name" in metadata
        ), "collection_name is required for DocumentRetrieval"

        if "chunked" not in metadata:
            metadata["chunked"] = False

        for obj in objects:
            obj["collection_name"] = metadata["collection_name"]
            if "chunk_spans" not in obj:
                obj["chunk_spans"] = []

        Retrieval.__init__(
            self,
            objects,
            payload_type="document",
            metadata=metadata,
            mapping=mapping,
            unmapped_keys=[
                "uuid",
                "ELYSIA_SUMMARY",
                "collection_name",
                "chunk_spans",
                "chunk_uuid",
                "_REF_ID",
            ],
            **kwargs,
        )
        self.async_init_completed = False

    async def async_init(self, client_manager: ClientManager) -> None:
        if not self.async_init_completed:
            await self._get_related_documents(client_manager)

    async def _get_related_documents(self, client_manager: ClientManager) -> None:
        """
        Get the related full documents for the chunked documents in the DocumentRetrieval object.
        """
        if self.metadata["chunked"]:
            chunked_collection_name = (
                f"ELYSIA_CHUNKED_{self.metadata['collection_name'].lower()}__"
            )
            async with client_manager.connect_to_async_client() as client:
                if await client.collections.exists(chunked_collection_name):
                    chunked_collection = client.collections.get(chunked_collection_name)

                    # re-retrieve the exact same items but with the full document references
                    chunked_response = (
                        await chunked_collection.query.fetch_objects_by_ids(
                            [object["uuid"] for object in self.objects],
                            return_references=QueryReference(link_on="fullDocument"),
                        )
                    )

                    # Loop over all references in this new response and add them to the full_docs dictionary
                    full_docs = {}
                    for object in chunked_response.objects:
                        # retrieve cross references from this current chunked response object
                        if "fullDocument" in object.references:
                            references = object.references["fullDocument"].objects

                            # this list is length one (only one full doc per chunk)
                            for full_document in references:

                                # each chunk is attached to a full doc, but can have multiple chunks per full doc
                                if str(full_document.uuid) not in full_docs:
                                    full_docs[str(full_document.uuid)] = {
                                        **full_document.properties,
                                        "uuid": str(full_document.uuid),
                                        "collection_name": self.metadata[
                                            "collection_name"
                                        ],
                                        "chunk_spans": [],
                                    }

                                chunk_spans: list[int] = object.properties["chunk_spans"]  # type: ignore

                                # chunk comes from `object` which is the chunk (outer level loop)
                                full_docs[str(full_document.uuid)][
                                    "chunk_spans"
                                ].append(
                                    {
                                        "start": chunk_spans[0],
                                        "end": chunk_spans[1],
                                        "uuid": str(object.uuid),
                                    }
                                )

                    self.full_documents = list(full_docs.values())
                else:
                    self.full_documents = self.objects
        else:
            self.full_documents = self.objects
        self.async_init_completed = True

    def full_documents_to_json(self, mapping: bool = False) -> list[dict]:
        assert all(
            isinstance(obj, dict) for obj in self.full_documents
        ), "All objects must be dictionaries"

        if mapping and self.mapping is not None:
            output_objects = []
            for obj in self.full_documents:
                output_objects.append(
                    {
                        key: obj[self.mapping[key]]
                        for key in self.mapping
                        if self.mapping[key] != ""
                    }
                )
                output_objects[-1].update(
                    {key: obj[key] for key in self.unmapped_keys if key in obj}
                )
        else:
            output_objects = self.full_documents

        for object in output_objects:
            format_dict_to_serialisable(object)

        return output_objects

    async def to_frontend(
        self, user_id: str, conversation_id: str, query_id: str
    ) -> dict:
        """
        Needs re-definition of the to_frontend method because the frontend receives full documents,
        whereas the LLM receives the chunks.
        So to_json() outputs the chunks (to the LLM)
        But the to_frontend() outputs the full documents and the chunk spans.
        """
        if self.metadata["chunked"] and not self.async_init_completed:
            raise Exception(
                "ERROR: Documents are chunked and DocumentRetrieval not initialized. "
                "Need to run .async_init(client_manager) to retrieve the full documents attached to the chunks. "
                "Otherwise set metadata['chunked'] to False"
            )

        payload = {
            "type": self.payload_type,
            "objects": (
                self.full_documents_to_json(mapping=True)
                if self.metadata["chunked"] and self.async_init_completed
                else self.to_json(mapping=True)
            ),
            "metadata": self.metadata,
        }

        if "code" in self.metadata:
            payload["code"] = self.metadata["code"]

        return {
            "type": self.frontend_type,
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "id": self.frontend_type[:3] + "-" + str(uuid.uuid4()),
            "payload": payload,
        }


class Aggregation(Retrieval):
    def __init__(
        self,
        objects: list[dict],
        metadata: dict = {},
        name: str | None = None,
        **kwargs,
    ) -> None:
        if name is None and "collection_name" in metadata:
            name = metadata["collection_name"]
        elif name is None:
            name = "default"

        Retrieval.__init__(
            self,
            objects=objects,
            payload_type="aggregation",
            metadata=metadata,
            name=name,
            **kwargs,
        )

    def llm_parse(self) -> str:
        out = ""
        if "collection_name" in self.metadata:
            out += f"\nAggregated collection: '{self.metadata['collection_name']}'"

            count = {}
            if len(self.objects) > 0:
                for metric in self.objects[0]["collections"][0][
                    self.metadata["collection_name"]
                ]:
                    if metric != "ELYSIA_NUM_ITEMS":
                        num_values = len(
                            self.objects[0]["collections"][0][
                                self.metadata["collection_name"]
                            ][metric]["values"]
                        )
                        count[metric] = num_values

                for metric in count:
                    out += f", returned {count[metric]} values for {metric}."
                if any(count[metric] == 0 for metric in count):
                    out += f" For those metrics with 0 objects, judge the aggregation code that was created, and evaluate whether it was appropriate for the collection/metric, the user prompt, and the data available."
                    out += f" If it seemed innappropriate, you can choose to try again if you think it can still be completed (or there is more to do)."
            else:
                out += f" but no objects were returned."
                out += f" Since it had no objects, judge the aggregation code that was created, and evaluate whether it was appropriate for the collection, the user prompt, and the data available."
                out += f" If it seemed innappropriate, you can choose to try again if you think it can still be completed (or there is more to do)."

            if "groupby_name" in self.metadata:
                out += f"\nGrouped by property '{self.metadata['groupby_name']}'"
            if "metrics" in self.metadata:
                out += f"\nReturned metrics for property(-ies) '{self.metadata['metrics']}'"
            if "impossible" in self.metadata:
                if "collection_name" in self.metadata:
                    out += f"\nThis attempt at aggregating the collection: {self.metadata['collection_name']} was deemed impossible."
                out += f"The user prompt was: '{self.metadata['impossible']}', which the aggregation model deemed impossible to complete."
                if "impossible_reasoning" in self.metadata:
                    out += f"\nReasoning for impossibility: {self.metadata['impossible_reasoning']}"
            if "aggregation_output" in self.metadata:
                out += f"\nThe aggregation query used was:\n{self.metadata['aggregation_output']}"
        return out



from elysia.tree.tree import Tree
from elysia.tree.util import get_saved_trees_weaviate

from elysia.util.client import ClientManager

import pytest


def test_save_load_local():

    tree = Tree(
        user_id="test_save_load_user",
        conversation_id="test_save_load_conversation",
        low_memory=True,
        style="This is a test style!",
        agent_description="This is a test agent description!",
        end_goal="This is a test end goal!",
    )
    tree.tree_data.environment.hidden_environment["Example Entry"] = (
        "This is an example!"
    )

    # save the tree to a file
    tree_json = tree.export_to_json()

    # load the tree from the file
    loaded_tree = Tree.import_from_json(tree_json)

    assert tree.user_id == loaded_tree.user_id
    assert tree.conversation_id == loaded_tree.conversation_id
    assert tree.low_memory == loaded_tree.low_memory
    assert tree.tree_data.atlas.style == loaded_tree.tree_data.atlas.style
    assert (
        tree.tree_data.atlas.agent_description
        == loaded_tree.tree_data.atlas.agent_description
    )
    assert tree.tree_data.atlas.end_goal == loaded_tree.tree_data.atlas.end_goal
    assert "Example Entry" in loaded_tree.tree_data.environment.hidden_environment
    assert (
        tree.tree_data.environment.hidden_environment["Example Entry"]
        == loaded_tree.tree_data.environment.hidden_environment["Example Entry"]
    )



import os

import pytest

from elysia.util.client import ClientManager


@pytest.mark.asyncio
async def test_client_manager_starts():
    try:
        client_manager = ClientManager()
    finally:
        await client_manager.close_clients()


@pytest.mark.asyncio
async def test_client_manager_starts_with_env_vars():
    try:
        client_manager = ClientManager()
        assert client_manager.wcd_url == os.getenv("WCD_URL")
        assert client_manager.wcd_api_key == os.getenv("WCD_API_KEY")
    finally:
        await client_manager.close_clients()


@pytest.mark.asyncio
async def test_client_manager_starts_with_api_keys():
    try:
        client_manager = ClientManager()

        if "OPENAI_API_KEY" in os.environ:
            assert "X-OpenAI-Api-Key" in client_manager.headers
            assert client_manager.headers["X-OpenAI-Api-Key"] == os.getenv(
                "OPENAI_API_KEY"
            )
    finally:
        await client_manager.close_clients()


@pytest.mark.asyncio
async def test_clients_are_connected():
    try:
        client_manager = ClientManager()
        await client_manager.start_clients()
        assert client_manager.client.is_ready()
    finally:
        await client_manager.close_clients()


@pytest.mark.asyncio
async def test_sync_client_connects():
    try:
        client_manager = ClientManager()
        with client_manager.connect_to_client() as client:
            assert client_manager.sync_in_use_counter == 1
        assert client_manager.sync_in_use_counter == 0
    finally:
        await client_manager.close_clients()


@pytest.mark.asyncio
async def test_async_client_connects():
    try:
        client_manager = ClientManager()
        async with client_manager.connect_to_async_client() as client:
            assert client_manager.async_in_use_counter == 1
            pass
        assert client_manager.async_in_use_counter == 0
    finally:
        await client_manager.close_clients()



from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse

from elysia.api.api_types import (
    ViewPaginatedCollectionData,
    UpdateCollectionMetadataData,
)
from elysia.api.core.log import logger
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from elysia.util.parsing import format_dict_to_serialisable
from elysia.util.collection import (
    async_get_collection_data_types,
    paginated_collection,
)
from elysia.preprocessing.collection import (
    edit_preprocessed_collection_async,
    delete_preprocessed_collection_async,
    preprocessed_collection_exists_async,
)
from elysia.util.return_types import specific_return_types, types_dict, all_return_types
from elysia.util.client import ClientManager

from weaviate.classes.query import Filter


router = APIRouter()


@router.get("/mapping_types")
async def mapping_types():
    """
    Retrieve mapping types from hardcoded values.

    Returns:
        (JSONResponse): A JSON response containing the mapping types.
    """
    logger.debug(f"/mapping_types API request received")

    headers = {"Cache-Control": "no-cache"}
    try:
        return JSONResponse(
            content={
                "mapping_types": [
                    {
                        "name": return_type,
                        "description": all_return_types[return_type],
                        "fields": types_dict[return_type],
                    }
                    for return_type in all_return_types
                ],
                "error": "",
            },
            status_code=200,
            headers=headers,
        )
    except Exception as e:
        logger.exception(f"Error in /mapping_types API")
        return JSONResponse(
            content={"mapping_types": [], "error": str(e)},
            status_code=500,
            headers=headers,
        )


@router.get("/{user_id}/list")
async def collections_list(
    user_id: str, user_manager: UserManager = Depends(get_user_manager)
):
    """
    Retrieve a list of collections from the currently connected Weaviate cluster for the user.

    Args:
        user_id (str): The ID of the user to retrieve collections for.
        user_manager (UserManager): The user manager.

    Returns:
        (JSONResponse): A JSON response containing:
            - collections (list[dict]): A list of collections with their:
                - name (str): The name of the collection.
                - total (int): The total number of objects in the collection.
                - vectorizer (dict): The vectoriser configuration for the collection.
                - processed (bool): Whether the collection has been processed.
                - error (bool): Whether there is an error retrieving _this specific_ collection.
            - error (str): An overall error message if there is an error in the main function, otherwise an empty string.
    """
    logger.debug(f"/collections API request received")

    headers = {"Cache-Control": "no-cache"}

    try:

        user_local = await user_manager.get_user_local(user_id)
        client_manager: ClientManager = user_local["client_manager"]

        if not client_manager.is_client:
            return JSONResponse(
                content={
                    "collections": [],
                    "error": (
                        "Client manager is not connected. "
                        "This is not an error if the user does not want to use Weaviate."
                    ),
                },
                status_code=200,
                headers=headers,
            )

        async with client_manager.connect_to_async_client() as client:

            collections = [
                c
                for c in await client.collections.list_all()
                if not c.startswith("ELYSIA_")
            ]

            # get processed collections
            if await client.collections.exists("ELYSIA_METADATA__"):
                metadata_collection = client.collections.get("ELYSIA_METADATA__")
                processed_collections = await metadata_collection.query.fetch_objects(
                    filters=Filter.any_of(
                        [Filter.by_property("name").equal(c) for c in collections]
                    ),
                    limit=9999,
                )
                processed_collection_names = [
                    processed_collection.properties["name"]
                    for processed_collection in processed_collections.objects
                ]
                processed_collections_prompts = {
                    processed_collection.properties[
                        "name"
                    ]: processed_collection.properties["prompts"]
                    for processed_collection in processed_collections.objects
                }
            else:
                processed_collection_names = []
                processed_collections_prompts = {}

            # get collection metadata
            metadata = []
            for collection_name in collections:
                logger.debug(
                    f"Gathering information for collection_name: {collection_name}"
                )

                try:
                    # for vectoriser
                    collection = client.collections.get(collection_name)
                    config = await collection.config.get()

                    # for total count
                    all_aggregate = await collection.aggregate.over_all(
                        total_count=True
                    )
                    len = all_aggregate.total_count

                    # for processed
                    processed = collection_name in processed_collection_names

                    if processed:
                        prompts = processed_collections_prompts[collection_name]
                    else:
                        prompts = []

                    vector_config = {"fields": {}, "global": {}}
                    if config.vector_config:
                        for (
                            named_vector_name,
                            named_vector_config,
                        ) in config.vector_config.items():

                            # for this vectoriser, what model is being used?
                            model = named_vector_config.vectorizer.model
                            if isinstance(model, dict) and "model" in model:
                                model = model["model"]
                            else:
                                model = "Unknown"

                            # check what fields are being vectorised
                            fields = named_vector_config.vectorizer.source_properties
                            if fields is None:
                                fields = [
                                    c.name
                                    for c in config.properties
                                    if c.data_type[:].startswith("text")
                                ]

                            # for each field, add the vectoriser and model to the vector config
                            for field_name in fields:
                                if field_name not in vector_config["fields"]:
                                    vector_config["fields"][field_name] = [
                                        {
                                            "named_vector": named_vector_name,
                                            "vectorizer": named_vector_config.vectorizer.vectorizer.name,
                                            "model": model,
                                        }
                                    ]
                                else:
                                    vector_config["fields"][field_name].append(
                                        {
                                            "named_vector": named_vector_name,
                                            "vectorizer": named_vector_config.vectorizer.vectorizer.name,
                                            "model": model,
                                        }
                                    )

                    if config.vectorizer_config:
                        model = config.vectorizer_config.model
                        if isinstance(model, dict) and "model" in model:
                            model = model["model"]

                        vector_config["global"] = {
                            "vectorizer": config.vectorizer_config.vectorizer.name,
                            "model": model,
                        }
                    else:
                        vector_config["global"] = {}

                    metadata.append(
                        {
                            "name": collection_name,
                            "total": len,
                            "vectorizer": vector_config,
                            "processed": processed,
                            "error": False,
                            "prompts": prompts,
                        }
                    )
                except Exception as e:
                    metadata.append(
                        {
                            "name": collection_name,
                            "total": 0,
                            "vectorizer": {},
                            "processed": False,
                            "error": True,
                            "prompts": [],
                        }
                    )

            return JSONResponse(
                content={"collections": metadata, "error": ""},
                status_code=200,
                headers=headers,
            )

    except Exception as e:
        logger.exception(f"Error in /collections API")
        return JSONResponse(
            content={"collections": [], "error": str(e)},
            status_code=500,
            headers=headers,
        )


@router.post("/{user_id}/view/{collection_name}")
async def view_paginated_collection(
    user_id: str,
    collection_name: str,
    data: ViewPaginatedCollectionData,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Find a list of objects in a collection, with pagination and filtering.

    Args:
        user_id (str): The ID of the user whose collections to view.
        collection_name (str): The name of the collection to view.
        data (ViewPaginatedCollectionData): The data for the request, containing:
            - page_size (int): The number of objects to return per page.
            - page_number (int): The page number to return.
            - query (str): The query to search for. If empty, all objects will be returned.
                If non-empty, BM25 will be used to search for objects.
            - sort_on (str): The property to sort on.
            - ascending (bool): Whether to sort in ascending or descending order.
            - filter_config (dict): The filter configuration, containing:
                - type (str): The type of filter to apply, one of 'all' or 'any'.
                - filters (list[dict]): The filters to apply, containing:
                    - field (str): The field to filter on.
                    - operator (str): The operator to apply,
                        one of 'equal', 'not_equal', 'greater_than', 'less_than', 'greater_than_or_equal',
                        'less_than_or_equal', 'contains', 'not_contains', 'is_empty', 'is_not_empty'.
                    - value (str): The value to filter on.
        user_manager (UserManager): The user manager.

    Returns:
        (JSONResponse): A JSON response containing:
            - properties (list[dict]): The properties of the collection.
            - items (list[dict]): The items in the collection.
    """

    logger.debug(f"/view_paginated_collection API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Query: {data.query}")
    logger.debug(f"Collection name: {collection_name}")
    logger.debug(f"Page size: {data.page_size}")
    logger.debug(f"Page number: {data.page_number}")
    logger.debug(f"Sort on: {data.sort_on}")
    logger.debug(f"Ascending: {data.ascending}")
    logger.debug(f"Filter config: {data.filter_config}")

    try:
        user_local = await user_manager.get_user_local(user_id)
        client_manager = user_local["client_manager"]
        async with client_manager.connect_to_async_client() as client:
            # get collection properties
            data_types = await async_get_collection_data_types(client, collection_name)

            # obtain paginated results from collection
            items = await paginated_collection(
                client=client,
                collection_name=collection_name,
                query=data.query,
                page_size=data.page_size,
                page_number=data.page_number,
                sort_on=data.sort_on,
                ascending=data.ascending,
                filter_config=data.filter_config,
            )

            logger.info(f"Returning collection info for {collection_name}")
            return JSONResponse(
                content={"properties": data_types, "items": items, "error": ""},
                status_code=200,
            )
    except Exception as e:
        logger.exception(f"Error in /view_paginated_collection API")
        return JSONResponse(
            content={"properties": [], "items": [], "error": str(e)}, status_code=500
        )


@router.get("/{user_id}/get_object/{collection_name}/{uuid}")
async def get_object(
    user_id: str,
    collection_name: str,
    uuid: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Get a single object from a collection by its UUID.

    Args:
        user_id (str): The ID of the user whose object to retrieve.
        collection_name (str): The name of the collection to retrieve the object from.
        uuid (str): The UUID of the object to retrieve.
        user_manager (UserManager): The user manager.

    Returns:
        (JSONResponse): A JSON response containing:
            - properties (list[dict]): The data types of the object.
            - items (list[dict]): The object itself (one element list).
            - error (str): An error message if there is an error, otherwise an empty string.
    """
    logger.debug(f"/get_object API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Collection name: {collection_name}")
    logger.debug(f"UUID: {uuid}")

    headers = {"Cache-Control": "max-age=300"}  # cache for 5 minutes

    try:
        user_local = await user_manager.get_user_local(user_id)
        client_manager = user_local["client_manager"]
        async with client_manager.connect_to_async_client() as client:
            collection = client.collections.get(collection_name)

            try:
                object = await collection.query.fetch_object_by_id(uuid)
                object = object.properties
            except Exception as e:
                logger.error(
                    f"Error in /get_object API: No object found with UUID {uuid} in collection {collection_name}"
                )
                return JSONResponse(
                    content={
                        "properties": [],
                        "items": [],
                        "error": f"No object found with UUID {uuid} in collection {collection_name}.",
                    },
                    status_code=404,
                    headers=headers,
                )

            data_types = await async_get_collection_data_types(client, collection_name)
            format_dict_to_serialisable(object)

    except Exception as e:
        logger.exception(f"Error in /get_object API")
        return JSONResponse(
            content={"properties": [], "items": [], "error": str(e)},
            status_code=500,
            headers=headers,
        )

    return JSONResponse(
        content={"properties": data_types, "items": [object], "error": ""},
        status_code=200,
        headers=headers,
    )


@router.get("/{user_id}/metadata/{collection_name}")
async def collection_metadata(
    user_id: str,
    collection_name: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Get the preprocessed metadata for a collection.
    This retrieves the output of the `.preprocess` function in Elysia.

    Args:
        user_id (str): The ID of the user whose metadata to retrieve.
        collection_name (str): The name of the collection to retrieve the metadata for.
        user_manager (UserManager): The user manager.

    Returns:
        (JSONResponse): A JSON response containing:
            - metadata (dict): The metadata for the collection.
            - error (str): An error message if there is an error, otherwise an empty string.

    Example metadata:
    ```json
    {
        "metadata": dict = {

            # summary statistics of each field in the collection
            "fields": list = [
                {
                    "name": str,
                    "description": str,
                    "range": list[float],
                    "type": str,
                    "groups": dict[str, str],
                    "mean": float
                },
                field_name_2: dict,
                ...
            ],

            # mapping_1, mapping_2 etc refer to frontend-specific types that the AI has deemed appropriate for this data
            # then the dict is to map the frontend fields to the data fields
            "mappings": dict = {
                mapping_1: dict,
                mapping_2: dict,
                ...,
            },

            # number of items in collection (float but just for consistency)
            "length": float,

            # AI generated summary of the dataset
            "summary": str,

            # name of collection
            "name": str,

            # what named vectors are available (if any)
            "named_vectors": list = [
                {
                    "name": str,
                    "vectorizer": str,
                    "model": str,
                    "enabled": bool,
                    "source_properties": list,
                    "description": str # defaults to empty
                },
                ...
            ],

            # vectoriser configuration
            "vectoriser": dict = {
                "vectorizer": str,
                "model": str,
            },

            # some config settings relevant for queries
            "index_properties": {
                "isNullIndexed": bool,
                "isLengthIndexed": bool,
                "isTimestampIndexed": bool,
            }
        }
        "error": ""
    }
    ```
    """
    logger.debug(f"/collection_metadata API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Collection name: {collection_name}")

    headers = {"Cache-Control": "no-cache"}

    try:
        user_local = await user_manager.get_user_local(user_id)
        client_manager = user_local["client_manager"]

        async with client_manager.connect_to_async_client() as client:
            metadata_name = f"ELYSIA_METADATA__"

            # check if the collection itself exists
            if not await client.collections.exists(collection_name):
                raise Exception(f"Collection {collection_name} does not exist")

            # check if the metadata collection exists
            if not await preprocessed_collection_exists_async(
                collection_name, client_manager
            ):
                raise Exception(
                    f"Metadata collection for {collection_name} does not exist"
                )

            metadata_collection = client.collections.get(metadata_name)
            metadata = await metadata_collection.query.fetch_objects(
                filters=Filter.by_property("name").equal(collection_name),
                limit=1,
            )
            properties = metadata.objects[0].properties
            format_dict_to_serialisable(properties)

            if properties["named_vectors"] is None:
                properties["named_vectors"] = []

    except Exception as e:
        logger.exception(f"Error in /collection_metadata API")
        return JSONResponse(
            content={
                "metadata": {},
                "error": str(e),
            },
            status_code=200,
            headers=headers,
        )

    logger.debug(f"\n\nProperties: {properties}\n\n")

    return JSONResponse(
        content={
            "metadata": properties,
            "error": "",
        },
        status_code=200,
        headers=headers,
    )


@router.patch("/{user_id}/metadata/{collection_name}")
async def update_metadata(
    user_id: str,
    collection_name: str,
    data: UpdateCollectionMetadataData,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Update the preprocessed metadata for a collection.
    This updates the output of the `.preprocess` function in Elysia.
    Any fields not provided will not be updated.

    Args:
        user_id (str): The ID of the user whose metadata to update.
        collection_name (str): The name of the collection to update the metadata for.
        data (UpdateCollectionMetadataData): The data for the request, containing:
            - named_vectors (list[dict]): The named vectors to update.
            - summary (str): The summary to update.
            - mappings (dict): The mappings to update.
            - fields (list[dict]): The fields to update.
        user_manager (UserManager): The user manager.

    Returns:
        (JSONResponse): A JSON response containing:
            - metadata (dict): The updated metadata for the collection (same as the "metadata" output of `collection_metadata`).
            - error (str): An error message if there is an error, otherwise an empty string.
    """
    logger.debug(f"/update_metadata API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Collection name: {collection_name}")
    logger.debug(f"Data: {data}")

    # retrieve the current metadata
    try:
        user_local = await user_manager.get_user_local(user_id)
        client_manager = user_local["client_manager"]

        properties = await edit_preprocessed_collection_async(
            collection_name=collection_name,
            client_manager=client_manager,
            named_vectors=(
                [
                    named_vector.model_dump()
                    for named_vector in data.named_vectors
                    if named_vector
                ]
                if data.named_vectors
                else None
            ),
            summary=data.summary,
            mappings=data.mappings,
            fields=(
                [field.model_dump() if field else None for field in data.fields]
                if data.fields
                else None
            ),
        )
    except Exception as e:
        logger.exception(f"Error in /update_metadata API:")
        return JSONResponse(
            content={"metadata": {}, "error": str(e)},
            status_code=200,
        )

    return JSONResponse(
        content={"metadata": properties, "error": ""},
        status_code=200,
    )


@router.delete("/{user_id}/metadata/{collection_name}")
async def delete_metadata(
    user_id: str,
    collection_name: str,
    user_manager: UserManager = Depends(get_user_manager),
) -> JSONResponse:
    """
    Delete the metadata for a collection.

    Args:
        user_id (str): The ID of the user.
        collection_name (str): The name of the collection.
        user_manager (UserManager): The user manager.

    Returns:
        (JSONResponse): A JSON response containing an error message if there is an error, otherwise an empty string.
    """
    logger.debug(f"/delete_metadata API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Collection name: {collection_name}")

    # retrieve the current metadata
    try:
        user_local = await user_manager.get_user_local(user_id)
        client_manager = user_local["client_manager"]

        await delete_preprocessed_collection_async(collection_name, client_manager)

    except Exception as e:
        logger.exception(f"Error in /delete_metadata API")
        return JSONResponse(content={"error": str(e)}, status_code=200)

    return JSONResponse(content={"error": ""}, status_code=200)


@router.delete("/{user_id}/metadata/delete/all")
async def delete_all_metadata(
    user_id: str,
    user_manager: UserManager = Depends(get_user_manager),
) -> JSONResponse:
    """
    Delete the metadata for a collection.

    Args:
        user_id (str): The ID of the user.
        user_manager (UserManager): The user manager.

    Returns:
        (JSONResponse): A JSON response containing an error message if there is an error, otherwise an empty string.
    """
    logger.debug(f"/delete_all_metadata API request received")
    logger.debug(f"User ID: {user_id}")

    # retrieve the current metadata
    try:
        user_local = await user_manager.get_user_local(user_id)
        client_manager: ClientManager = user_local["client_manager"]
        async with client_manager.connect_to_async_client() as client:
            if await client.collections.exists("ELYSIA_METADATA__"):
                await client.collections.delete("ELYSIA_METADATA__")

    except Exception as e:
        logger.exception(f"Error in /delete_all_metadata API")
        return JSONResponse(content={"error": str(e)}, status_code=200)

    return JSONResponse(content={"error": ""}, status_code=200)



import asyncio
import pytest
from elysia.util.dummy_adapter import DummyAdapter
from dspy import configure, ChatAdapter
import dspy


@pytest.fixture(scope="session")
def event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    try:
        yield loop
    finally:
        loop.close()


@pytest.fixture(autouse=True, scope="module")
def use_dummy_adapter():

    prev_adapter = dspy.settings.adapter
    configure(adapter=DummyAdapter())
    yield
    configure(adapter=prev_adapter)



import datetime
import json
import uuid
from typing import Any

from weaviate.collections.classes.aggregate import (
    AggregateDate,
    AggregateGroupByReturn,
    AggregateNumber,
    AggregateReturn,
    AggregateText,
)


def objects_dict_to_str(objects: list) -> str:
    out = ""
    for item in objects:
        if isinstance(item, dict):
            inner = "["
            for key, value in item.items():
                try:
                    inner += json.dumps({key: value}) + ", "
                except TypeError:
                    pass
            inner += "]"
            out += inner + "\n"

        elif isinstance(item, list):
            out += objects_dict_to_str(item)

        elif isinstance(item, str):
            out += item + "\n"

    return out


def format_datetime(dt: datetime.datetime | None) -> str:
    if dt is None:
        return ""

    output = dt.isoformat("T")
    plus = output.find("+")
    if plus != -1:
        return output[:plus] + "Z"
    else:
        return output + "Z"


def format_dict_to_serialisable(d: dict[str, Any], remove_unserialisable: bool = False):
    if remove_unserialisable:
        keys_to_remove = []

    for key, value in d.items():
        if isinstance(value, datetime.datetime):
            d[key] = format_datetime(value)

        elif isinstance(value, dict):
            format_dict_to_serialisable(value, remove_unserialisable)

        elif isinstance(value, list):
            items_to_remove = []
            for i, item in enumerate(value):
                if isinstance(item, dict):
                    format_dict_to_serialisable(item, remove_unserialisable)

                elif isinstance(item, datetime.datetime):
                    d[key][i] = format_datetime(item)

                elif isinstance(item, uuid.UUID):
                    d[key][i] = str(item)

                elif remove_unserialisable and not isinstance(
                    item, (str, int, float, bool, list, dict)
                ):
                    items_to_remove.append(i)

            for index in sorted(items_to_remove, reverse=True):
                del d[key][index]

        elif isinstance(value, uuid.UUID):
            d[key] = str(value)

        elif remove_unserialisable and not isinstance(
            d, (str, int, float, bool, list, dict)
        ):
            keys_to_remove.append(key)

    if remove_unserialisable:
        for key in keys_to_remove:
            del d[key]


def remove_whitespace(text: str) -> str:
    return " ".join(text.split())


def format_aggregation_property(prop):
    if isinstance(prop, AggregateText):
        out = {"type": "text", "values": []}
        for top_occurence in prop.top_occurrences:
            out["values"].append(
                {
                    "value": top_occurence.count,
                    "field": top_occurence.value,
                    "aggregation": "count",
                }
            )
        return out

    elif isinstance(prop, AggregateNumber):
        out = {"type": "number", "values": []}

        if prop.count is not None:
            out["values"].append(
                {"value": prop.count, "field": None, "aggregation": "count"}
            )

        if prop.maximum is not None:
            out["values"].append(
                {"value": prop.maximum, "field": None, "aggregation": "maximum"}
            )

        if prop.mean is not None:
            out["values"].append(
                {"value": prop.mean, "field": None, "aggregation": "mean"}
            )

        if prop.median is not None:
            out["values"].append(
                {"value": prop.median, "field": None, "aggregation": "median"}
            )

        if prop.minimum is not None:
            out["values"].append(
                {"value": prop.minimum, "field": None, "aggregation": "minimum"}
            )

        if prop.mode is not None:
            out["values"].append(
                {"value": prop.mode, "field": None, "aggregation": "mode"}
            )

        if prop.sum_ is not None:
            out["values"].append(
                {"value": prop.sum_, "field": None, "aggregation": "sum"}
            )

        return out

    elif isinstance(prop, AggregateDate):
        out = {"type": "date", "values": []}

        if prop.count is not None:
            out["values"].append(
                {"value": prop.count, "field": None, "aggregation": "count"}
            )

        if prop.maximum is not None:
            out["values"].append(
                {"value": prop.maximum, "field": None, "aggregation": "maximum"}
            )

        if prop.median is not None:
            out["values"].append(
                {"value": prop.median, "field": None, "aggregation": "median"}
            )

        if prop.minimum is not None:
            out["values"].append(
                {"value": prop.minimum, "field": None, "aggregation": "minimum"}
            )

        if prop.mode is not None:
            out["values"].append(
                {"value": prop.mode, "field": None, "aggregation": "mode"}
            )

        return out
    else:
        return {"type": "unknown", "values": []}


def format_aggregation_response(response):
    out = {}
    if isinstance(response, AggregateGroupByReturn):

        for result in response.groups:
            field = result.grouped_by.prop
            out[field] = {"type": "text", "values": []}

        for result in response.groups:
            field = result.grouped_by.prop

            if result.total_count is not None:
                out[field]["values"].append(
                    {
                        "value": result.total_count,
                        "field": result.grouped_by.value,
                        "aggregation": "count",
                    }
                )

            for key, prop in result.properties.items():
                formatted_props = format_aggregation_property(prop)
                if "groups" not in out[field]:
                    out[field]["groups"] = {}

                if result.grouped_by.value not in out[field]["groups"]:
                    out[field]["groups"][result.grouped_by.value] = {}

                out[field]["groups"][result.grouped_by.value][key] = formatted_props

    elif isinstance(response, AggregateReturn):

        for field, field_properties in response.properties.items():
            out[field] = {}
            props = format_aggregation_property(field_properties)
            for key, prop in props.items():
                out[field][key] = prop

    return out



import asyncio
import os
import pytest
from typing import Any
from weaviate.classes.query import Filter, QueryReference

from elysia.tree.objects import Environment
from elysia.objects import (
    Completed,
    Response,
    Status,
    Text,
    Update,
    Warning,
    Result,
)
from elysia.tools.retrieval.chunk import AsyncCollectionChunker
from elysia.tools.retrieval.objects import (
    ConversationRetrieval,
    DocumentRetrieval,
    MessageRetrieval,
)
from elysia.util.client import ClientManager
from elysia.util.parsing import format_dict_to_serialisable


async def start_client_manager():
    client_manager = ClientManager(
        wcd_url=os.getenv("WCD_URL"),
        wcd_api_key=os.getenv("WCD_API_KEY"),
        openai_api_key=os.getenv("OPENAI_API_KEY"),
    )
    await client_manager.start_clients()
    return client_manager


async def _retrieve_unchunked_documents(
    collection_name: str, client_manager: ClientManager
):
    async with client_manager.connect_to_async_client() as client:
        collection = client.collections.get(collection_name)
        documents = await collection.query.fetch_objects(
            limit=2,
            return_references=QueryReference(link_on="isChunked"),
        )
    return documents


async def _retrieve_chunked_documents(
    collection_name: str, client_manager: ClientManager
):
    async with client_manager.connect_to_async_client() as client:
        collection = client.collections.get(collection_name)
        documents = await collection.query.fetch_objects(
            limit=2,
            return_references=QueryReference(link_on="fullDocument"),
        )
    return documents


async def _create_chunked_reference(
    collection_name: str, client_manager: ClientManager
):
    collection_chunker = AsyncCollectionChunker(collection_name)
    await collection_chunker.create_chunked_reference("content", client_manager)


async def _chunk_documents(
    collection_name: str, response, client_manager: ClientManager
):
    collection_chunker = AsyncCollectionChunker(collection_name)
    await collection_chunker(response, "content", client_manager)


async def get_collection_information(
    collection_name: str, client_manager: ClientManager
):
    async with client_manager.connect_to_async_client() as client:
        metadata_name = "ELYSIA_METADATA__"
        metadata_collection = client.collections.get(metadata_name)
        metadata = await metadata_collection.query.fetch_objects(
            filters=Filter.by_property("name").equal(collection_name),
            limit=1,
        )
        properties = metadata.objects[0].properties
        format_dict_to_serialisable(properties)
        return properties


async def get_document_retrievals():
    collection_name = "Ml_wikipedia"
    client_manager = await start_client_manager()

    # get the collection information
    collection_information = await get_collection_information(
        collection_name, client_manager
    )
    await _create_chunked_reference(collection_name, client_manager)

    # retrieve some documents
    unchunked_documents = await _retrieve_unchunked_documents(
        collection_name, client_manager
    )

    # chunk the documents
    await _chunk_documents(collection_name, unchunked_documents, client_manager)

    # retrieve the chunked documents
    chunked_documents = await _retrieve_chunked_documents(
        f"ELYSIA_CHUNKED_{collection_name.lower()}__", client_manager
    )

    unchunked_document_properties = [
        doc.properties for doc in unchunked_documents.objects
    ]
    chunked_document_properties = [doc.properties for doc in chunked_documents.objects]

    for i, prop in enumerate(unchunked_document_properties):
        prop["uuid"] = str(unchunked_documents.objects[i].uuid)

    for i, prop in enumerate(chunked_document_properties):
        prop["uuid"] = str(chunked_documents.objects[i].uuid)

    # create a retrieval object
    non_chunked_retrieval = DocumentRetrieval(
        objects=unchunked_document_properties,
        metadata={"collection_name": "ml_wikipedia", "chunked": False},
        mapping=collection_information["mappings"]["document"],
    )
    chunked_retrieval = DocumentRetrieval(
        objects=chunked_document_properties,
        metadata={"collection_name": "ml_wikipedia", "chunked": True},
        mapping=collection_information["mappings"]["document"],
    )
    await chunked_retrieval.async_init(client_manager)
    await client_manager.close_clients()

    return non_chunked_retrieval, chunked_retrieval


@pytest.mark.asyncio
async def test_document_retrieval():
    client_manager = await start_client_manager()

    # if the collection is not found, skip the test
    if not client_manager.client.collections.exists("Ml_wikipedia"):
        pytest.skip("Collection not found in Weaviate")

    non_chunked_retrieval, chunked_retrieval = await get_document_retrievals()

    chunked_json = chunked_retrieval.to_json(mapping=False)

    chunked_fe = await chunked_retrieval.to_frontend(
        "user_id", "conversation_id", "query_id"
    )

    # Check that the document attached to the chunk is >= length of the chunk
    for i, full_doc_obj in enumerate(chunked_fe["payload"]["objects"]):
        assert len(full_doc_obj["content"]) >= len(chunked_json[i]["content"])

    # Check that all chunk spans from the chunk exist in the full documnet
    all_chunk_spans = [
        (obj["uuid"], obj["chunk_spans"])
        for obj in chunked_json
        if obj["chunk_spans"] != []
    ]

    # Find all unique instances of chunkID+chunkSpan
    which_chunk_spans_exist = {
        str(all_chunk_spans[i][0]) + str(all_chunk_spans[i][1]): False
        for i in range(len(all_chunk_spans))
    }

    # Look through full documents (FE payload-objects) and check if the chunk content is present
    for i, full_doc_obj in enumerate(chunked_fe["payload"]["objects"]):
        for chunk_span in full_doc_obj["chunk_spans"]:
            ind = str(chunk_span["uuid"]) + str(
                [chunk_span["start"], chunk_span["end"]]
            )
            if ind in which_chunk_spans_exist:
                which_chunk_spans_exist[ind] = True

    assert all(which_chunk_spans_exist.values())



"""
More complex testing of retrieval accuracy.
I.e. testing whether the correct data is retrieved for a given prompt.
This also includes testing the constructed query code contains the correct fields/filters etc.
"""

# TODO



import pytest
from fastapi.responses import JSONResponse
import os
import json

from elysia.api.core.log import set_log_level
from elysia.api.dependencies.common import get_user_manager

set_log_level("CRITICAL")

from elysia.api.routes.init import initialise_tree, initialise_user
from elysia.api.api_types import InitialiseTreeData


def read_response(response: JSONResponse):
    return json.loads(response.body)


class TestTree:

    @pytest.mark.asyncio
    async def test_initialise_user(self):
        user_manager = get_user_manager()
        user_id = "test_new_user"

        # Initialise when a user does not exist
        out = await initialise_user(
            user_id,
            user_manager,
        )

        response = read_response(out)
        assert response["error"] == ""
        assert response["user_exists"] is False

        # check config options are environment variables
        if "BASE_MODEL" in os.environ:
            assert response["config"]["settings"]["BASE_MODEL"] == os.getenv(
                "BASE_MODEL"
            )
        if "OPENAI_API_KEY" in os.environ:
            assert response["config"]["settings"]["API_KEYS"][
                "openai_api_key"
            ] == os.getenv("OPENAI_API_KEY")

        # Initialise when a user exists
        out = await initialise_user(
            user_id,
            user_manager,
        )

        response = read_response(out)
        assert response["error"] == ""
        assert response["user_exists"] is True

    @pytest.mark.asyncio
    async def test_tree(self):

        user_manager = get_user_manager()
        user_id = "test_user"
        conversation_id = "test_conversation"

        out = await initialise_user(
            user_id,
            user_manager,
        )
        response = read_response(out)
        assert response["error"] == ""

        out = await initialise_tree(
            user_id,
            conversation_id,
            InitialiseTreeData(
                low_memory=True,
            ),
            user_manager,
        )

        response = read_response(out)
        assert response["error"] == ""

        out = await initialise_tree(
            user_id + "2",
            conversation_id + "2",
            InitialiseTreeData(
                low_memory=True,
            ),
            user_manager,
        )

        response = read_response(out)
        assert response["error"] != ""  # should error as user doesn't exist



from httpx import delete


def test_create_tools_simple():

    from elysia import tool

    @tool
    async def add1(x: int, y: int) -> int:
        """
        Return the sum of two numbers.
        """
        return x + y

    from elysia import Tree

    tree = Tree()
    tree.add_tool(add1)

    response, objects = tree.run("What is 1123 + 48332?")

    assert str(1123 + 48332) in response

    tree = Tree()

    @tool(tree=tree, branch_id="base")
    async def add2(x: int, y: int) -> int:
        return x + y

    @tool
    async def calculate_two_numbers(x: int, y: int):
        """
        This function calculates the sum, product, and difference of two numbers.
        """
        yield {
            "sum": x + y,
            "product": x * y,
            "difference": x - y,
        }
        yield f"I just performed some calculations on {x} and {y}."

    tree = Tree()
    tree.add_tool(calculate_two_numbers)

    response, objects = tree.run("What is 1123 + 48332?")

    env = tree.tree_data.environment.environment
    assert "calculate_two_numbers" in env
    assert "default" in env["calculate_two_numbers"]

    result = env["calculate_two_numbers"]["default"][0]
    assert "sum" in result["objects"][0]
    assert result["objects"][0]["sum"] == 1123 + 48332
    assert result["objects"][0]["product"] == 1123 * 48332
    assert result["objects"][0]["difference"] == 1123 - 48332

    from math import prod
    from elysia import Error

    @tool
    async def perform_mathematical_operations(
        numbers: list[int | float], operation: str = "sum"
    ):
        """
        This function calculates a mathematical operation on the `numbers` list.
        The `numbers` input must be a list of integers or floats.
        The `operation` input must be one of: "sum" or "product". These are the only options.
        """

        if operation == "sum":
            yield sum(numbers)
        elif operation == "product":
            yield prod(numbers)
            # This will return an error back to the decision tree
            yield Error(
                f"You picked the input {operation}, but it was not in the available operations: 'sum' or 'product'"
            )
            return  # Then return out of the tool early

        yield f"I just performed a {operation} on {numbers}."

    tree = Tree()
    tree.add_tool(perform_mathematical_operations)

    tree("What is 2379 x 234 x 213 x 3?")

    env = tree.tree_data.environment.environment
    result = env["perform_mathematical_operations"]["default"][0]
    assert result["objects"][0]["tool_result"] == 2379 * 234 * 213 * 3


def test_query_weaviate():

    try:
        from rich import print

        import requests, json

        url = "https://raw.githubusercontent.com/weaviate/weaviate-examples/main/jeopardy_small_dataset/jeopardy_tiny.json"
        resp = requests.get(url)
        data = json.loads(resp.text)

        from elysia.util.client import ClientManager
        from weaviate.classes.config import Configure

        client_manager = ClientManager()

        with client_manager.connect_to_client() as client:

            if client.collections.exists("JeopardyQuestion"):
                client.collections.delete("JeopardyQuestion")

            client.collections.create(
                "JeopardyQuestion",
                vector_config=Configure.Vectors.text2vec_weaviate(),
            )

            jeopardy = client.collections.get("JeopardyQuestion")
            response = jeopardy.data.insert_many(data)

            if response.has_errors:
                print(response.errors)
            else:
                print("Insert complete.")

        from elysia import preprocess

        preprocess("JeopardyQuestion", force=True)

        from elysia import view_preprocessed_collection

        print(view_preprocessed_collection("JeopardyQuestion"))

        from elysia import Tree

        tree = Tree()

        print(tree.view())

        tree.tree

        response, objects = tree(
            "Find a single question about Science",
            collection_names=["JeopardyQuestion"],
        )

        tree("What about animals?")

        print(tree.conversation_history)

    finally:
        client_manager = ClientManager()

        with client_manager.connect_to_client() as client:
            if client.collections.exists("JeopardyQuestion"):
                client.collections.delete("JeopardyQuestion")

            from elysia import delete_preprocessed_collection

            delete_preprocessed_collection("JeopardyQuestion", client_manager)



import time
import uuid
from copy import deepcopy
import dspy
from pydantic import BaseModel
from typing import Any
from elysia.util.parsing import format_dict_to_serialisable
from logging import Logger
from elysia.objects import Update


class Tracker:
    """
    Simple class to track:
    - the average time taken for an LLM call
    - number of calls made
    - number of input/output tokens used
    """

    def __init__(self, tracker_names: list[str], logger: Logger):
        self.trackers = {
            name: {
                "timer": {
                    "calls": 0,
                    "avg_time": 0,
                    "total_time": 0,
                    "start_time": None,
                    "end_time": None,
                }
            }
            for name in tracker_names
        }
        self.trackers = {
            **self.trackers,
            "models": {
                "base_lm": {
                    "calls": 0,
                    "cost": None,
                    "avg_cost": None,
                    "input_tokens": None,
                    "output_tokens": None,
                    "avg_input_tokens": None,
                    "avg_output_tokens": None,
                },
                "complex_lm": {
                    "calls": 0,
                    "cost": None,
                    "avg_cost": None,
                    "input_tokens": None,
                    "output_tokens": None,
                    "avg_input_tokens": None,
                    "avg_output_tokens": None,
                },
            },
        }
        self.logger = logger

    def start_tracking(self, tracker_name: str):
        self.trackers[tracker_name]["timer"]["start_time"] = time.perf_counter()

    def update_lm_costs(self, lm: dspy.LM | None = None, model_type: str = "base_lm"):

        if lm is not None:

            # check how many calls have been made
            prev_calls = self.trackers["models"][model_type]["calls"]
            total_calls = len(lm.history)
            self.trackers["models"][model_type]["calls"] = total_calls
            num_calls = total_calls - prev_calls

            if num_calls == 0:
                return

            history = lm.history[-num_calls:]

            input_tokens = 0
            for h in history:
                if "usage" in h and "prompt_tokens" in h["usage"]:
                    input_tokens += h["usage"]["prompt_tokens"]

            output_tokens = 0
            for h in history:
                if "usage" in h and "completion_tokens" in h["usage"]:
                    output_tokens += h["usage"]["completion_tokens"]

            cost = 0
            for h in history:
                if "cost" in h and h["cost"] is not None:
                    cost += h["cost"]

            if self.trackers["models"][model_type]["input_tokens"] is None:
                self.trackers["models"][model_type]["input_tokens"] = input_tokens
            else:
                self.trackers["models"][model_type]["input_tokens"] += input_tokens

            if self.trackers["models"][model_type]["output_tokens"] is None:
                self.trackers["models"][model_type]["output_tokens"] = output_tokens
            else:
                self.trackers["models"][model_type]["output_tokens"] += output_tokens

            if self.trackers["models"][model_type]["cost"] is None:
                self.trackers["models"][model_type]["cost"] = cost
            else:
                self.trackers["models"][model_type]["cost"] += cost

    def end_tracking(
        self,
        tracker_name: str,
        call_name: str = "",
        base_lm: dspy.LM | None = None,
        complex_lm: dspy.LM | None = None,
    ):
        if self.trackers[tracker_name]["timer"]["start_time"] is None:
            self.logger.warning(f"Tracker {tracker_name} has not been started yet!")
            return

        self.trackers[tracker_name]["timer"]["calls"] += 1

        time_taken = (
            time.perf_counter() - self.trackers[tracker_name]["timer"]["start_time"]
        )
        self.update_avg_time(tracker_name, time_taken)
        self.update_lm_costs(base_lm, "base_lm")
        self.update_lm_costs(complex_lm, "complex_lm")

        if call_name != "":
            self.logger.debug(
                f"Time taken for {call_name} ({tracker_name}): {time_taken: .2f} seconds"
            )
        else:
            self.logger.debug(
                f"Time taken for {tracker_name}: {time_taken: .2f} seconds"
            )

    def update_avg_time(self, tracker_name: str, time_taken: float):
        self.trackers[tracker_name]["timer"]["total_time"] += time_taken
        self.trackers[tracker_name]["timer"]["avg_time"] = (
            self.trackers[tracker_name]["timer"]["total_time"]
            / self.trackers[tracker_name]["timer"]["calls"]
        )

    def remove_tracker(self, tracker_name: str):
        self.trackers.pop(tracker_name)

    def add_tracker(self, tracker_name: str):
        self.trackers[tracker_name] = {
            "timer": {
                "calls": 0,
                "avg_time": 0,
                "total_time": 0,
                "start_time": None,
                "end_time": None,
            }
        }

    def get_num_calls(self, model_type: str):
        return self.trackers["models"][model_type]["calls"]

    def get_average_time(self, tracker_name: str):
        return self.trackers[tracker_name]["timer"]["avg_time"]

    def get_total_input_tokens(self, model_type: str):
        return self.trackers["models"][model_type]["input_tokens"]

    def get_total_output_tokens(self, model_type: str):
        return self.trackers["models"][model_type]["output_tokens"]

    def get_total_cost(self, model_type: str):
        return self.trackers["models"][model_type]["cost"]

    def get_average_input_tokens(self, model_type: str):
        return (
            (
                self.trackers["models"][model_type]["input_tokens"]
                / self.trackers["models"][model_type]["calls"]
            )
            if self.trackers["models"][model_type]["input_tokens"] is not None
            else 0
        )

    def get_average_output_tokens(self, model_type: str):
        return (
            (
                self.trackers["models"][model_type]["output_tokens"]
                / self.trackers["models"][model_type]["calls"]
            )
            if self.trackers["models"][model_type]["output_tokens"] is not None
            else 0
        )

    def get_average_cost(self, model_type: str):
        return (
            (
                self.trackers["models"][model_type]["cost"]
                / self.trackers["models"][model_type]["calls"]
            )
            if self.trackers["models"][model_type]["cost"] is not None
            else 0
        )

    def reset_trackers(self):
        for tracker in self.trackers:
            self.trackers[tracker]["timer"] = {
                "calls": 0,
                "avg_time": 0,
                "total_time": 0,
                "start_time": None,
                "end_time": None,
            }
            self.trackers[tracker]["models"] = {
                "base_lm": {
                    "calls": 0,
                    "input": None,
                    "output": None,
                },
                "complex_lm": {
                    "calls": 0,
                    "input": None,
                    "output": None,
                },
            }


class TreeUpdate:
    """
    Frontend update to represent what nodes have been updated.
    """

    def __init__(
        self,
        from_node: str,
        to_node: str,
        reasoning: str,
        reset_tree: bool = False,
    ):
        """
        Args:
            from_node (str): The node that is being updated from.
            to_node (str): The node that is being updated to.
            reasoning (str): The reasoning for the update.
            last_in_branch (bool): Whether this is the last update in the branch (whether the tree is complete after this - hardcoded)
                e.g. in query tool, sometimes the query is the end of the tree and sometimes summarise objects is
        """
        self.from_node = from_node
        self.to_node = to_node
        self.reasoning = reasoning
        self.reset_tree = reset_tree

    async def to_frontend(
        self,
        user_id: str,
        conversation_id: str,
        query_id: str,
        tree_index: int,
    ):
        return {
            "type": "tree_update",
            "id": str(uuid.uuid4()),
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "payload": {
                "node": self.from_node,
                "tree_index": tree_index,
                "decision": self.to_node,
                "reasoning": self.reasoning,
                "reset": self.reset_tree,
            },
        }


class TrainingUpdate:
    """
    Record a training example for a module.
    Keep track of the inputs and outputs of the module, and the module name.
    """

    def __init__(
        self,
        module_name: str,
        inputs: dict,
        outputs: dict,
        extra_inputs: dict = {},
    ):
        self.module_name = module_name

        # Format datetime in inputs and outputs
        format_dict_to_serialisable(inputs)
        format_dict_to_serialisable(outputs)

        # Check if any Pydantic base models exist in either inputs or outputs
        # Create copies of inputs and outputs to avoid modifying the original dictionaries
        inputs_copy = deepcopy(inputs)
        outputs_copy = deepcopy(outputs)

        # If so, convert them to a dictionary
        for key, value in inputs_copy.items():
            inputs_copy[key] = self._convert_basemodel(value)
        for key, value in outputs_copy.items():
            outputs_copy[key] = self._convert_basemodel(value)

        self.inputs = {**inputs_copy, **extra_inputs}
        self.outputs = outputs_copy

    def _convert_basemodel(self, value: Any):
        if isinstance(value, BaseModel):
            return value.model_dump()
        elif isinstance(value, dict):
            return {k: self._convert_basemodel(v) for k, v in value.items()}
        elif isinstance(value, list):
            return [self._convert_basemodel(v) for v in value]
        else:
            return value

    def to_json(self):
        return {
            "module_name": self.module_name,
            "inputs": self.inputs,
            "outputs": self.outputs,
        }


class FewShotExamples(Update):
    """
    A set of few shot examples for a module.
    """

    def __init__(self, uuids: list[str]):
        self.uuids = uuids
        Update.__init__(
            self,
            "fewshot_examples",
            {
                "uuids": self.uuids,
            },
        )



import os
import pytest

from elysia.config import Settings
from elysia.api.routes.init import initialise_user
from elysia.api.routes.user_config import (
    load_a_config,
    save_config_user,
    load_config_user,
    new_user_config,
    get_current_user_config,
)
from elysia.api.api_types import SaveConfigUserData
from fastapi.responses import JSONResponse
import json
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from weaviate.util import generate_uuid5

from uuid import uuid4


def read_response(response: JSONResponse):
    return json.loads(response.body)


@pytest.mark.asyncio
async def test_set_api_keys():

    user_id = f"test_{uuid4()}"
    user_manager = get_user_manager()

    openrouter_api_key = "some-random-key"

    # initialise the user
    response = await initialise_user(user_id=user_id, user_manager=user_manager)
    response = read_response(response)

    # create a new config
    response = await new_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    config_id = response["config"]["id"]

    new_config = response["config"]
    new_config["settings"]["API_KEYS"]["openrouter_api_key"] = openrouter_api_key

    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=response["config"]["name"],
            default=True,
            config=new_config,
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert (
        response["config"]["settings"]["API_KEYS"]["openrouter_api_key"]
        == openrouter_api_key
    )

    # load a config
    response = await get_current_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert (
        response["config"]["settings"]["API_KEYS"]["openrouter_api_key"]
        == openrouter_api_key
    )


@pytest.mark.asyncio
async def test_change_models():

    user_id = "test_change_models"
    user_manager = get_user_manager()

    # initialise the user
    response = await initialise_user(user_id=user_id, user_manager=user_manager)
    response = read_response(response)

    # create a new config
    response = await new_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    config_id = response["config"]["id"]

    # change config, change the models
    new_config = response["config"]
    new_config["settings"]["BASE_MODEL"] = "gpt-4o"
    new_config["settings"]["COMPLEX_MODEL"] = "gpt-4o"
    new_config["settings"]["BASE_PROVIDER"] = "openrouter/openai"
    new_config["settings"]["COMPLEX_PROVIDER"] = "openrouter/openai"

    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=response["config"]["name"],
            default=True,
            config=new_config,
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert response["config"]["settings"]["BASE_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["COMPLEX_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["BASE_PROVIDER"] == "openrouter/openai"
    assert response["config"]["settings"]["COMPLEX_PROVIDER"] == "openrouter/openai"

    # load a config
    response = await get_current_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert response["config"]["settings"]["BASE_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["COMPLEX_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["BASE_PROVIDER"] == "openrouter/openai"
    assert response["config"]["settings"]["COMPLEX_PROVIDER"] == "openrouter/openai"



from elysia.tree.tree import Tree
from elysia.tree.util import get_saved_trees_weaviate

from elysia.util.client import ClientManager

import pytest


@pytest.mark.asyncio
async def test_save_load_weaviate():

    try:
        tree = Tree(
            user_id="test_save_load_user",
            conversation_id="test_save_load_conversation",
            low_memory=True,
            style="This is a test style!",
            agent_description="This is a test agent description!",
            end_goal="This is a test end goal!",
        )
        tree.tree_data.environment.hidden_environment["Example Entry"] = (
            "This is an example!"
        )

        client_manager = ClientManager()

        # save the tree to weaviate
        await tree.export_to_weaviate(
            collection_name="Test_ELYSIA_save_load_collection",
            client_manager=client_manager,
        )

        # load the tree from weaviate
        loaded_tree = await Tree.import_from_weaviate(
            collection_name="Test_ELYSIA_save_load_collection",
            conversation_id=tree.conversation_id,
            client_manager=client_manager,
        )

        assert tree.user_id == loaded_tree.user_id
        assert tree.conversation_id == loaded_tree.conversation_id
        assert tree.low_memory == loaded_tree.low_memory
        assert tree.tree_data.atlas.style == loaded_tree.tree_data.atlas.style
        assert (
            tree.tree_data.atlas.agent_description
            == loaded_tree.tree_data.atlas.agent_description
        )
        assert tree.tree_data.atlas.end_goal == loaded_tree.tree_data.atlas.end_goal
        assert "Example Entry" in loaded_tree.tree_data.environment.hidden_environment
        assert (
            tree.tree_data.environment.hidden_environment["Example Entry"]
            == loaded_tree.tree_data.environment.hidden_environment["Example Entry"]
        )

        saved_trees = await get_saved_trees_weaviate(
            collection_name="Test_ELYSIA_save_load_collection",
            client_manager=client_manager,
        )

        assert tree.conversation_id in saved_trees

    finally:
        if client_manager.client.collections.exists("Test_ELYSIA_save_load_collection"):
            client_manager.client.collections.delete("Test_ELYSIA_save_load_collection")

        await client_manager.close_clients()



import pytest
import elysia
from elysia.tree.tree import Tree
from elysia.util.client import ClientManager
from elysia.preprocessing import preprocess
from weaviate.util import generate_uuid5
import weaviate.classes as wvc


@pytest.fixture(scope="module")
def create_and_process_collection():
    collection_name = "Test_ELYSIA_Product_Data"
    client_manager = ClientManager()
    create_regular_vectorizer_collection(client_manager, collection_name)
    preprocess(collection_name, client_manager)
    yield
    with client_manager.connect_to_client() as client:
        if client.collections.exists(collection_name):
            client.collections.delete(collection_name)
        client_manager.close_clients()


def create_regular_vectorizer_collection(
    client_manager: ClientManager, collection_name: str
):

    with client_manager.connect_to_client() as client:

        if client.collections.exists(collection_name):
            client.collections.delete(collection_name)

        collection = client.collections.create(
            name=collection_name,
            properties=[
                wvc.config.Property(
                    name="product_id", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(
                    name="description", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(
                    name="tags", data_type=wvc.config.DataType.TEXT_ARRAY
                ),
                wvc.config.Property(name="price", data_type=wvc.config.DataType.NUMBER),
            ],
            vectorizer_config=wvc.config.Configure.Vectorizer.text2vec_openai(
                vectorize_collection_name=False
            ),
        )

        products_data = [
            {
                "product_id": "prod1",
                "title": "Wireless Bluetooth Headphones",
                "description": "High-quality wireless headphones with noise cancellation and 30-hour battery life.",
                "tags": ["electronics", "audio", "wireless", "bluetooth"],
                "price": 199.99,
            },
            {
                "product_id": "prod2",
                "title": "Organic Green Tea",
                "description": "Premium organic green tea sourced from high-altitude tea gardens in Japan.",
                "tags": ["beverage", "tea", "organic", "healthy"],
                "price": 24.99,
            },
            {
                "product_id": "prod3",
                "title": "Yoga Mat with Alignment Lines",
                "description": "Non-slip yoga mat with helpful alignment lines for proper posture during practice.",
                "tags": ["fitness", "yoga", "exercise", "wellness"],
                "price": 39.99,
            },
            {
                "product_id": "prod4",
                "title": "Smart Home Security Camera",
                "description": "WiFi-enabled security camera with motion detection and smartphone notifications.",
                "tags": ["security", "smart-home", "camera", "surveillance"],
                "price": 149.99,
            },
        ]

        with collection.batch.fixed_size(batch_size=10) as batch:
            for product in products_data:
                batch.add_object(
                    properties={
                        "title": product["title"],
                        "description": product["description"],
                        "tags": product["tags"],
                        "price": product["price"],
                    },
                    uuid=generate_uuid5(product["product_id"]),
                )


def test_various_methods(create_and_process_collection):
    collection_name = "Test_ELYSIA_Product_Data"
    tree = Tree(
        low_memory=False,
    )
    user_prompt = "Find me some bluetooth headphones"
    _, _ = tree(
        user_prompt,
        collection_names=[
            collection_name,
        ],
    )

    # check the user prompt is in the conversation history
    assert tree.tree_data.conversation_history[0]["content"] == user_prompt
    assert tree.tree_data.conversation_history[0]["role"] == "user"

    # check some variables are set
    assert tree.tree_data.num_trees_completed > 0
    assert tree.tree_data.user_prompt == user_prompt

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)
    assert len(all_decision_history) > 0
    assert "query" in all_decision_history

    # this should have returned something in the environment
    assert "query" in tree.tree_data.environment.environment
    for env_name in tree.tree_data.environment.environment["query"]:
        for returned_object in tree.tree_data.environment.environment["query"][
            env_name
        ]:
            assert "metadata" in returned_object
            assert len(returned_object["objects"]) > 0

    # this should have updated tasks_completed
    assert len(tree.tree_data.tasks_completed) > 0
    prompt_found = False
    for task in tree.tree_data.tasks_completed:
        if task["prompt"] == user_prompt:
            prompt_found = True
            assert len(task["task"]) > 0
            break

    assert prompt_found

    # check the tree resets
    tree.soft_reset()

    # These should not change
    assert len(tree.tree_data.conversation_history) > 0
    assert len(tree.tree_data.tasks_completed) > 0
    assert (
        "query" in tree.tree_data.environment.environment
    )  # environment should not be reset

    # these should be reset
    assert tree.tree_data.num_trees_completed == 0
    assert tree.tree_index == 1

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)
    assert len(all_decision_history) == 0


def test_get_follow_up_suggestions(
    create_and_process_collection,
):  # Added fixture parameter
    tree = Tree()
    _, _ = tree("Find me some bluetooth headphones")
    suggestions = tree.get_follow_up_suggestions(
        context="make it about headphones specifically",
        num_suggestions=5,
    )

    assert len(suggestions) == 5



import asyncio
import datetime
import os
from typing import Any
import uuid
from dotenv import load_dotenv
from weaviate.util import generate_uuid5

# Load environment variables from .env file
load_dotenv(override=True)

from elysia.tree.tree import Tree
from elysia.util.client import ClientManager
from elysia.tree.util import delete_tree_from_weaviate
from elysia.api.utils.config import Config, BranchInitType
from elysia.config import Settings


class TreeManager:
    """
    Manages trees (different conversations) for a single user.
    Designed to be used with the Elysia API, or a manager for multiple Elysia decision trees.

    You can initialise the TreeManager with particular config options.
    Or, upon adding a tree, you can set a specific style, description and end goal for that tree.

    Each tree has separate elements: the tree itself, the last request time, and the asyncio event.
    """

    def __init__(
        self,
        user_id: str,
        config: Config | None = None,
        tree_timeout: datetime.timedelta | int | None = None,
    ):
        """
        Args:
            user_id (str): Required. A unique identifier for the user being managed within this TreeManager.
            config (Config | None): Optional. A config for all trees managed by this TreeManager.
                Defaults to a new config with default settings.
                If `settings` is not provided, it will be set to the default settings (smart_setup).
            tree_timeout (datetime.timedelta | int | None): Optional. A timeout for all trees managed by this TreeManager.
                Defaults to the value of the `TREE_TIMEOUT` environment variable.
                If an integer is passed, it will be interpreted as minutes.
                If set to 0, trees will not be automatically removed.
        """
        self.trees = {}
        self.user_id = user_id

        if tree_timeout is None:
            self.tree_timeout = datetime.timedelta(
                minutes=int(os.environ.get("TREE_TIMEOUT", 10))
            )
        elif isinstance(tree_timeout, int):
            self.tree_timeout = datetime.timedelta(minutes=tree_timeout)
        else:
            self.tree_timeout = tree_timeout

        if config is None:
            self.config = Config()
        else:
            self.config = config

        self.settings = self.config.settings

    def update_config(
        self,
        conversation_id: str | None = None,
        config_id: str | None = None,
        config_name: str | None = None,
        settings: dict[str, Any] | None = None,
        style: str | None = None,
        agent_description: str | None = None,
        end_goal: str | None = None,
        branch_initialisation: BranchInitType | None = None,
    ):
        if config_id is not None:
            self.config.id = config_id

        if config_name is not None:
            self.config.name = config_name

        if settings is not None:
            self.configure(conversation_id=conversation_id, replace=True, **settings)

        if style is not None:
            self.change_style(style, conversation_id)

        if agent_description is not None:
            self.change_agent_description(agent_description, conversation_id)

        if end_goal is not None:
            self.change_end_goal(end_goal, conversation_id)

        if branch_initialisation is not None:
            self.change_branch_initialisation(branch_initialisation, conversation_id)

    def add_tree(
        self,
        conversation_id: str,
        low_memory: bool = False,
    ):
        """
        Add a tree to the TreeManager.
        The decision tree can be initialised with specific config options, such as style, agent description and end goal.
        As well as a Settings object, which chooses options such as the LLM models, API keys and more.

        Args:
            conversation_id (str): Required. A unique identifier for the conversation.
            low_memory (bool): Optional. Whether to use low memory mode for the tree.
                Controls the LM history being saved in the tree, and some other variables.
                Defaults to False.
        """

        if not self.tree_exists(conversation_id):
            self.trees[conversation_id] = {
                "tree": Tree(
                    conversation_id=conversation_id,
                    user_id=self.user_id,
                    settings=self.settings,
                    style=self.config.style,
                    agent_description=self.config.agent_description,
                    end_goal=self.config.end_goal,
                    branch_initialisation=self.config.branch_initialisation,
                    low_memory=low_memory,
                    use_elysia_collections=self.config.use_elysia_collections,
                ),
                "last_request": datetime.datetime.now(),
                "event": asyncio.Event(),
            }
            self.trees[conversation_id]["event"].set()

    async def save_tree_weaviate(
        self, conversation_id: str, client_manager: ClientManager
    ):
        """
        Save a tree to Weaviate to collection ELYSIA_TREES__.
        Creates the collection if it doesn't exist.

        Args:
            conversation_id (str): The conversation ID which contains the tree.
            client_manager (ClientManager): The client manager to use for the tree.
        """
        tree: Tree = self.get_tree(conversation_id)
        await tree.export_to_weaviate("ELYSIA_TREES__", client_manager)

    async def check_tree_exists_weaviate(
        self, conversation_id: str, client_manager: ClientManager
    ):
        """
        Check if a tree exists in a Weaviate instance.
        The collection ELYSIA_TREES__ must exist, returns False if it doesn't.

        Args:
            conversation_id (str): The conversation ID which contains the tree.
            client_manager (ClientManager): The client manager to use for the tree.

        Returns:
            (bool): True if the tree exists in the Weaviate instance, False otherwise.
        """
        async with client_manager.connect_to_async_client() as client:
            if not await client.collections.exists("ELYSIA_TREES__"):
                return False

            collection = client.collections.get("ELYSIA_TREES__")
            uuid = generate_uuid5(conversation_id)
            return await collection.data.exists(uuid)

    async def load_tree_weaviate(
        self, conversation_id: str, client_manager: ClientManager
    ):
        """
        Load a tree from Weaviate.
        The conversation ID from the loaded tree is placed into the tree manager
        (possibly overwriting an existing tree with the same conversation ID).
        Then the tree itself is not returned - instead the list of frontend payloads
        that were yielded to the frontend by the tree is returned.

        Args:
            conversation_id (str): The conversation ID which contains the tree.
            client_manager (ClientManager): The client manager to use for the tree.

        Returns:
            (list): A list of dictionaries, each containing a frontend payload that was used to generate the tree.
                The list is ordered by the time the payload was originally sent to the frontend (at the time it was saved).
        """
        tree = await Tree.import_from_weaviate(
            "ELYSIA_TREES__", conversation_id, client_manager
        )
        if conversation_id not in self.trees:
            self.trees[conversation_id] = {
                "tree": None,
                "event": asyncio.Event(),
                "last_request": datetime.datetime.now(),
            }
        self.trees[conversation_id]["tree"] = tree
        self.trees[conversation_id]["event"].set()
        self.update_tree_last_request(conversation_id)
        return tree.returner.store

    async def delete_tree_weaviate(
        self, conversation_id: str, client_manager: ClientManager
    ):
        """
        Delete a tree from the stored trees in Weaviate.

        Args:
            conversation_id (str): The conversation ID of the tree to be deleted.
            client_manager (ClientManager): The client manager pointing to the Weaviate instance containing the tree.
        """
        await delete_tree_from_weaviate(
            conversation_id, "ELYSIA_TREES__", client_manager
        )

    def delete_tree_local(self, conversation_id: str):
        """
        Delete a tree from the TreeManager.

        Args:
            conversation_id (str): The conversation ID of the tree to be deleted.
        """
        if conversation_id in self.trees:
            del self.trees[conversation_id]

    def tree_exists(self, conversation_id: str):
        """
        Check if a tree exists in the TreeManager.

        Args:
            conversation_id (str): The conversation ID which may contain the tree.

        Returns:
            (bool): True if the tree exists, False otherwise.
        """
        return conversation_id in self.trees

    def get_tree(self, conversation_id: str):
        """
        Get a tree from the TreeManager.
        Will raise a ValueError if the tree is not found.

        Args:
            conversation_id (str): The conversation ID which contains the tree.

        Returns:
            (Tree): The tree associated with the conversation ID.
        """
        if conversation_id not in self.trees:
            raise ValueError(
                f"Tree {conversation_id} not found. Please initialise a tree first (by calling `add_tree`)."
            )

        return self.trees[conversation_id]["tree"]

    def get_event(self, conversation_id: str):
        """
        Get the asyncio.Event for a tree in the TreeManager.
        This is cleared when the tree is processing, and set when the tree is idle.
        This is used to block the API from sending multiple requests to the tree at once.

        Args:
            conversation_id (str): The conversation ID which contains the tree.

        Returns:
            (asyncio.Event): The event for the tree.
        """
        return self.trees[conversation_id]["event"]

    def configure(
        self, conversation_id: str | None = None, replace: bool = False, **kwargs: Any
    ):
        """
        Configure the settings for a tree in the TreeManager.

        Args:
            conversation_id (str | None): The conversation ID which contains the tree.
            replace (bool): Whether to override the current settings with the new settings.
                When this is True, all existing settings are removed, and only the new settings are used.
                Defaults to False.
            **kwargs (Any): The keyword arguments to pass to the Settings.configure() method.
        """
        if conversation_id is None:
            self.settings.configure(replace=replace, **kwargs)
            self.config.settings = self.settings
        else:
            self.get_tree(conversation_id).settings.configure(replace=replace, **kwargs)

    def change_style(self, style: str, conversation_id: str | None = None):
        """
        Change the style for a tree in the TreeManager.
        Or change the global style for all trees (if conversation_id is not supplied).
        And applies these changes to current trees with default settings.

        Args:
            style (str): The new style for the tree.
            conversation_id (str | None): Optional. The conversation ID which contains the tree.
                If not supplied, the style will be changed on all trees.
        """
        if conversation_id is None:
            self.config.style = style
            for conversation_id in self.trees:
                if not self.trees[conversation_id]["tree"]._config_modified:
                    self.trees[conversation_id]["tree"].change_style(style)
                    self.trees[conversation_id]["tree"]._config_modified = False
        else:
            if conversation_id not in self.trees:
                raise ValueError(f"Tree {conversation_id} not found")

            self.trees[conversation_id]["tree"].change_style(style)

    def change_agent_description(
        self, agent_description: str, conversation_id: str | None = None
    ):
        """
        Change the agent description for a tree in the TreeManager.
        Or change the global agent description for all trees (if conversation_id is not supplied).
        And applies these changes to current trees with default settings.

        Args:
            agent_description (str): The new agent description for the tree.
            conversation_id (str | None): Optional. The conversation ID which contains the tree.
                If not supplied, the agent description will be changed on all trees.
        """
        if conversation_id is None:
            self.config.agent_description = agent_description
            for conversation_id in self.trees:
                if not self.trees[conversation_id]["tree"]._config_modified:
                    self.trees[conversation_id]["tree"].change_agent_description(
                        agent_description
                    )
                    self.trees[conversation_id]["tree"]._config_modified = False
        else:
            if conversation_id not in self.trees:
                raise ValueError(f"Tree {conversation_id} not found")

            self.trees[conversation_id]["tree"].change_agent_description(
                agent_description
            )

    def change_end_goal(self, end_goal: str, conversation_id: str | None = None):
        """
        Change the end goal for a tree in the TreeManager.
        Or change the global end goal for all trees (if conversation_id is not supplied).
        And applies these changes to current trees with default settings.

        Args:
            end_goal (str): The new end goal for the tree.
            conversation_id (str | None): Optional. The conversation ID which contains the tree.
                If not supplied, the end goal will be changed on all trees.
        """
        if conversation_id is None:
            self.config.end_goal = end_goal
            for conversation_id in self.trees:
                if not self.trees[conversation_id]["tree"]._config_modified:
                    self.trees[conversation_id]["tree"].change_end_goal(end_goal)
                    self.trees[conversation_id]["tree"]._config_modified = False
        else:
            if conversation_id not in self.trees:
                raise ValueError(f"Tree {conversation_id} not found")

            self.trees[conversation_id]["tree"].change_end_goal(end_goal)

    def change_branch_initialisation(
        self, branch_initialisation: BranchInitType, conversation_id: str | None = None
    ):
        """
        Change the branch initialisation for a tree in the TreeManager.
        Or change the global branch initialisation for all trees (if conversation_id is not supplied).
        And applies these changes to current trees with default settings.

        Args:
            branch_initialisation (str): The new branch initialisation for the tree.
            conversation_id (str | None): Optional. The conversation ID which contains the tree.
                If not supplied, the branch initialisation will be changed on all trees.
        """
        if conversation_id is None:
            self.config.branch_initialisation = branch_initialisation
            for conversation_id in self.trees:
                if not self.trees[conversation_id]["tree"]._config_modified:
                    self.trees[conversation_id]["tree"].set_branch_initialisation(
                        branch_initialisation
                    )
                    self.trees[conversation_id]["tree"]._config_modified = False
        else:
            if conversation_id not in self.trees:
                raise ValueError(f"Tree {conversation_id} not found")

            self.trees[conversation_id]["tree"].set_branch_initialisation(
                branch_initialisation
            )

    async def process_tree(
        self,
        query: str,
        conversation_id: str,
        query_id: str | None = None,
        training_route: str = "",
        collection_names: list[str] = [],
        client_manager: ClientManager | None = None,
    ):
        """
        Process a tree in the TreeManager.
        This is an async generator which yields results from the tree.async_run() method.

        Args:
            query (str): Required. The user input/prompt to process in the decision tree.
            conversation_id (str): Required. The conversation ID which contains the tree.
            query_id (str): Optional. A unique identifier for the query.
                If not supplied, a random UUID will be generated.
            training_route (str): Optional. The training route, a string of the form "tool1/tool2/tool1" etc.
                See the `tree.async_run()` method for more details.
            collection_names (list[str]): Optional. A list of collection names to use in the query.
                If not supplied, all collections will be used.
            client_manager (ClientManager | None): Optional. A client manager to use for the query.
                If not supplied, a new ClientManager will be created.
        """

        if query_id is None:
            query_id = str(uuid.uuid4())

        tree: Tree = self.get_tree(conversation_id)
        self.update_tree_last_request(conversation_id)

        # wait for the tree to be idle
        await self.trees[conversation_id]["event"].wait()

        # clear the event, set it to working
        self.trees[conversation_id]["event"].clear()

        try:
            async for yielded_result in tree.async_run(
                query,
                collection_names=collection_names,
                client_manager=client_manager,
                query_id=query_id,
                training_route=training_route,
                close_clients_after_completion=False,
            ):
                yield yielded_result
                self.update_tree_last_request(conversation_id)

        finally:
            # set the event to idle
            self.trees[conversation_id]["event"].set()

    def check_tree_timeout(self, conversation_id: str):
        """
        Check if a tree has been idle for the last tree_timeout.

        Args:
            conversation_id (str): The conversation ID which contains the tree.

        Returns:
            (bool): True if the tree has been idle for the last tree_timeout, False otherwise.
        """
        # if tree not found, return True
        if conversation_id not in self.trees:
            return True

        # Remove any trees that have not been active in the last TREE_TIMEOUT minutes
        if (
            "last_request" in self.trees[conversation_id]
            and datetime.datetime.now() - self.trees[conversation_id]["last_request"]
            > self.tree_timeout
        ):
            return True

        return False

    def update_tree_last_request(self, conversation_id: str):
        self.trees[conversation_id]["last_request"] = datetime.datetime.now()

    def check_all_trees_timeout(self):
        """
        Check all trees in the TreeManager and remove any that have not been active in the last tree_timeout.
        """
        if self.tree_timeout == datetime.timedelta(minutes=0):
            return

        convs_to_remove = []
        for i, conversation_id in enumerate(self.trees):
            if self.check_tree_timeout(conversation_id):
                convs_to_remove.append(conversation_id)

        for conversation_id in convs_to_remove:
            del self.trees[conversation_id]



import pytest
import os
import dotenv

from elysia.api.services.user import UserManager
from elysia.api.services.tree import TreeManager
from elysia.config import Settings
from elysia.api.utils.config import Config

dotenv.load_dotenv(override=True)

from uuid import uuid4


@pytest.mark.asyncio
async def test_user_manager():
    """
    Test basic user manager functionality.
    """
    user_id = f"test_{uuid4()}"
    conversation_id = f"test_{uuid4()}"

    user_manager = UserManager()

    # will raise error if no user created
    with pytest.raises(ValueError):
        await user_manager.get_user_local(user_id)

    # add a user
    await user_manager.add_user_local(user_id)

    # get user
    user = await user_manager.get_user_local(user_id)
    assert user is not None

    # will raise error if no tree created
    with pytest.raises(ValueError):
        await user_manager.get_tree(user_id, conversation_id)

    # add a tree
    await user_manager.initialise_tree(user_id, conversation_id, low_memory=True)

    # get tree
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tree is not None


@pytest.mark.asyncio
async def test_tree_manager():
    """
    Test basic tree manager functionality.
    """
    user_id = f"test_{uuid4()}"
    conversation_id = f"test_{uuid4()}"

    tree_manager = TreeManager(user_id)

    # will raise error if no tree created
    with pytest.raises(ValueError):
        tree_manager.get_tree(conversation_id)

    # add a tree
    tree_manager.add_tree(conversation_id, low_memory=True)

    # get tree
    tree = tree_manager.get_tree(conversation_id)
    assert tree is not None



from logging import Logger
import datetime
import os
from typing import Literal, Optional
from uuid import uuid4

from elysia.config import Settings
from elysia.util.client import ClientManager

BranchInitType = Literal["default", "one_branch", "multi_branch", "empty"]


class Config:

    def __init__(
        self,
        id: str | None = None,
        name: str | None = None,
        settings: Settings | None = None,
        style: str | None = None,
        agent_description: str | None = None,
        end_goal: str | None = None,
        branch_initialisation: BranchInitType = "one_branch",
        use_elysia_collections: bool = True,
    ):

        if id is None:
            self.id = str(uuid4())
        else:
            self.id = id

        if name is None:
            self.name = "New Config"
        else:
            self.name = name

        if settings is None:
            self.settings = Settings().from_smart_setup()
        else:
            self.settings = settings

        if style is None:
            self.style = "Informative, polite and friendly."
        else:
            self.style = style

        if agent_description is None:
            self.agent_description = "You search and query Weaviate to satisfy the user's query, providing a concise summary of the results."
        else:
            self.agent_description = agent_description

        if end_goal is None:
            self.end_goal = (
                "You have satisfied the user's query, and provided a concise summary of the results. "
                "Or, you have exhausted all options available, or asked the user for clarification."
            )
        else:
            self.end_goal = end_goal

        self.branch_initialisation: BranchInitType = branch_initialisation
        self.use_elysia_collections: bool = use_elysia_collections

    def to_json(self):
        return {
            "id": self.id,
            "name": self.name,
            "settings": self.settings.to_json(),
            "style": self.style,
            "agent_description": self.agent_description,
            "end_goal": self.end_goal,
            "branch_initialisation": (
                self.branch_initialisation
                if self.branch_initialisation is not None
                else "one_branch"
            ),
            "use_elysia_collections": self.use_elysia_collections,
        }

    @classmethod
    def from_json(cls, json: dict):
        if "id" not in json and "config_id" in json:
            json["id"] = json["config_id"]

        return cls(
            id=json["id"],
            name=json["name"],
            settings=Settings.from_json(json["settings"]),
            style=json["style"],
            agent_description=json["agent_description"],
            end_goal=json["end_goal"],
            branch_initialisation=json["branch_initialisation"],
            use_elysia_collections=(
                json["use_elysia_collections"]
                if "use_elysia_collections" in json
                else True
            ),
        )


class FrontendConfig:
    """
    Handles frontend-specific config options.
    Anything that does not fit into the Elysia backend/python package, but relevant to the app is here.
    """

    def __init__(self, logger: Logger):
        """
        Args:
            logger (Logger): The logger to use for logging.
        """

        self.logger = logger

        client_timeout = int(os.getenv("CLIENT_TIMEOUT", 3))
        tree_timeout = int(os.getenv("TREE_TIMEOUT", 10))

        self.config: dict = {  # default values
            "save_trees_to_weaviate": True,
            "save_configs_to_weaviate": True,
            "client_timeout": client_timeout,
            "tree_timeout": tree_timeout,
        }
        self.save_location_wcd_url: str = os.getenv("WCD_URL", "")
        self.save_location_wcd_api_key: str = os.getenv("WCD_API_KEY", "")
        self.save_location_client_manager: ClientManager = ClientManager(
            wcd_url=self.save_location_wcd_url,
            wcd_api_key=self.save_location_wcd_api_key,
            logger=logger,
            client_timeout=self.config["client_timeout"],
        )

    def update_save_location(self, wcd_url: str, wcd_api_key: str):
        self.save_location_wcd_url = wcd_url
        self.save_location_wcd_api_key = wcd_api_key

    def get_save_location(self):
        return self.save_location_wcd_url, self.save_location_wcd_api_key

    def to_json(self):
        return {
            "save_trees_to_weaviate": self.config["save_trees_to_weaviate"],
            "save_configs_to_weaviate": self.config["save_configs_to_weaviate"],
            "save_location_wcd_url": self.save_location_wcd_url,
            "save_location_wcd_api_key": self.save_location_wcd_api_key,
            "client_timeout": self.config["client_timeout"],
            "tree_timeout": self.config["tree_timeout"],
        }

    async def configure(self, **kwargs):
        """
        Args:
            **kwargs:
                - save_trees_to_weaviate (bool): Whether to save trees to Weaviate.
                - save_config_to_weaviate (bool): Whether to save config to Weaviate.
                - tree_timeout (int): Optional.
                    The length of time in minutes a tree can be idle before being timed out.
                    Defaults to 10 minutes or the value of the TREE_TIMEOUT environment variable (integer, minutes).
                    If an integer is provided, it is interpreted as the number of minutes.
                - client_timeout (int): Optional.
                    The length of time in minutes a client can be idle before being timed out.
                    Defaults to 3 minutes or the value of the CLIENT_TIMEOUT environment variable (integer, minutes).
                    If an integer is provided, it is interpreted as the number of minutes.
                - save_location_wcd_url (str): Optional.
                    The URL of the Weaviate database to save trees/configs to.
                    Defaults to the value of the WCD_URL environment variable.
                - save_location_wcd_api_key (str): Optional.
                    The API key for the Weaviate database to save trees/configs to.
                    Defaults to the value of the WCD_API_KEY environment variable.
        """

        reload_client_manager = False
        if "save_location_wcd_url" in kwargs:
            self.save_location_wcd_url = kwargs["save_location_wcd_url"]
            reload_client_manager = True
        if "save_location_wcd_api_key" in kwargs:
            self.save_location_wcd_api_key = kwargs["save_location_wcd_api_key"]
            reload_client_manager = True
        if "save_trees_to_weaviate" in kwargs:
            self.config["save_trees_to_weaviate"] = kwargs["save_trees_to_weaviate"]
        if "save_configs_to_weaviate" in kwargs:
            self.config["save_configs_to_weaviate"] = kwargs["save_configs_to_weaviate"]
        if "client_timeout" in kwargs:
            self.config["client_timeout"] = kwargs["client_timeout"]
            self.save_location_client_manager.client_timeout = datetime.timedelta(
                minutes=self.config["client_timeout"]
            )
        if "tree_timeout" in kwargs:
            self.config["tree_timeout"] = kwargs["tree_timeout"]

        if reload_client_manager:
            await self.save_location_client_manager.reset_keys(
                wcd_url=self.save_location_wcd_url,
                wcd_api_key=self.save_location_wcd_api_key,
            )

    @classmethod
    async def from_json(cls, json: dict, logger: Logger):
        fe_config = cls(logger=logger)
        await fe_config.configure(**json)

        return fe_config



import uuid


def error_payload(text: str, conversation_id: str, query_id: str):
    return {
        "type": "error",
        "id": f"err-{str(uuid.uuid4())}",
        "conversation_id": conversation_id,
        "query_id": query_id,
        "payload": {"text": text},
    }



import asyncio
import pytest
from fastapi.responses import JSONResponse

import json

from elysia.api.core.log import logger, set_log_level
from elysia.api.dependencies.common import get_user_manager

set_log_level("CRITICAL")

from elysia.api.routes.query import process
from elysia.api.api_types import QueryData, InitialiseTreeData, SaveConfigUserData
from elysia.api.routes.init import initialise_user, initialise_tree
from elysia.api.routes.user_config import save_config_user

from uuid import uuid4


def read_response(response: JSONResponse):
    return json.loads(response.body)


class fake_websocket:
    results = []

    async def send_json(self, data: dict):
        self.results.append(data)


async def initialise_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()

    response = await initialise_user(
        user_id,
        user_manager,
    )

    response = await initialise_tree(
        user_id,
        conversation_id,
        InitialiseTreeData(
            low_memory=True,
        ),
        user_manager,
    )


@pytest.mark.asyncio
async def test_query():
    """
    Test the query endpoint.
    """

    user_id = f"test_{uuid4()}"
    conversation_id = f"test_{uuid4()}"
    query_id = f"test_{uuid4()}"
    config_id = f"test_{uuid4()}"

    websocket = fake_websocket()

    await initialise_user_and_tree(user_id, conversation_id)

    # set config for models
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name="test_config",
            default=True,
            config={
                "settings": {
                    "BASE_MODEL": "gpt-4o-mini",
                    "BASE_PROVIDER": "openai",
                    "COMPLEX_MODEL": "gpt-4o",
                    "COMPLEX_PROVIDER": "openai",
                },
            },
            frontend_config={
                "save_trees_to_weaviate": False,
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=get_user_manager(),
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]

    out = await process(
        QueryData(
            user_id=user_id,
            conversation_id=conversation_id,
            query="hi!",
            query_id=query_id,
            collection_names=[
                "Test_ELYSIA_collection_1",
                "Test_ELYSIA_collection_2",
            ],
        ).model_dump(),
        websocket,
        get_user_manager(),
    )

    # check all payloads are valid
    ner_found = False
    title_found = False
    complete_found = False
    for i, result in enumerate(websocket.results):
        if result["type"] == "ner":
            ner_found = True
        if result["type"] == "title":
            title_found = True
        if result["type"] == "completed":
            complete_found = True

        assert result["type"] != "error", result["payload"]["text"]
        assert isinstance(result, dict)
        assert "type" in result
        assert "id" in result
        assert "conversation_id" in result
        assert "query_id" in result
        assert "payload" in result
        assert isinstance(result["payload"], dict)
        if "objects" in result["payload"]:
            for obj in result["payload"]["objects"]:
                assert isinstance(obj, dict)
        if "metadata" in result["payload"]:
            assert isinstance(result["payload"]["metadata"], dict)
        if "text" in result["payload"]:
            assert isinstance(result["payload"]["text"], str)

        if result["type"] == "ner":
            assert isinstance(result["payload"]["text"], str)
            assert isinstance(result["payload"]["entity_spans"], list)
            assert isinstance(result["payload"]["noun_spans"], list)
            assert isinstance(result["payload"]["error"], str)
            assert result["payload"]["error"] == ""

        if result["type"] == "title":
            assert isinstance(result["payload"]["title"], str)
            assert isinstance(result["payload"]["error"], str)
            assert result["payload"]["error"] == ""

        if result["type"] == "completed":
            assert i == len(websocket.results) - 1
            assert websocket.results[i - 1]["type"] == "title"

    assert ner_found
    assert title_found
    assert complete_found



from logging import Logger
from elysia.objects import Tool, Error
from elysia.util.elysia_chain_of_thought import ElysiaChainOfThought
from elysia.tools.visualisation.objects import (
    ChartResult,
    BarChart,
    HistogramChart,
    ScatterOrLineChart,
)
from elysia.tools.visualisation.util import convert_chart_types_to_matplotlib
from elysia.tools.visualisation.prompt_templates import (
    CreateBarChart,
    CreateHistogramChart,
    CreateScatterOrLineChart,
)
from elysia.objects import Response
from elysia.util.objects import TrainingUpdate, TreeUpdate


class Visualise(Tool):

    def __init__(self, logger: Logger | None = None, **kwargs):
        super().__init__(
            name="visualise",
            description=(
                "Visualise data in a chart from the environment. "
                "You can only visualise data that is in the environment. "
                "If there is nothing relevant in the environment, do not choose this tool."
            ),
            status="Visualising...",
            end=True,
            inputs={
                "chart_type": {
                    "type": str,
                    "description": (
                        "The type of chart to create. "
                        "Must be one of: `bar`, `histogram`, `scatter_or_line`."
                    ),
                    "required": True,
                    "default": "",
                }
            },
        )
        self.logger = logger

    async def __call__(
        self,
        tree_data,
        inputs: dict,
        base_lm,
        complex_lm,
        client_manager,
        **kwargs,
    ):
        chart_type = inputs["chart_type"]

        try:
            type_mapping = {
                "bar": CreateBarChart,
                "histogram": CreateHistogramChart,
                "scatter_or_line": CreateScatterOrLineChart,
            }

            create_chart = ElysiaChainOfThought(
                type_mapping[chart_type],
                tree_data=tree_data,
                environment=True,
                impossible=True,
                tasks_completed=True,
                reasoning=tree_data.settings.BASE_USE_REASONING,
            )
            prediction = await create_chart.aforward(lm=base_lm)
            charts = prediction.charts
            title = prediction.overall_title

            yield Response(text=prediction.message_update)
            yield TrainingUpdate(
                module_name="visualise",
                inputs={
                    "chart_type": chart_type,
                    **tree_data.to_json(),
                },
                outputs=prediction.__dict__["_store"],
            )

            if prediction.impossible:
                yield ChartResult(
                    [],
                    chart_type,
                    title,
                    metadata={
                        "impossible": True,
                        "impossible_reasoning": (
                            prediction.reasoning
                            if tree_data.settings.BASE_USE_REASONING
                            else ""
                        ),
                    },
                )
                return

            if self.logger and self.logger.level <= 20:
                fig = convert_chart_types_to_matplotlib(charts)
                fig.show()

            yield ChartResult(charts, chart_type, title)
        except Exception as e:
            yield Error(f"Error visualising data: {str(e)}")

    async def is_tool_available(self, tree_data, base_lm, complex_lm, client_manager):
        """
        Available when there are relevant objects in the environment.
        """
        return not tree_data.environment.is_empty()



import asyncio
import pytest
from elysia.util.dummy_adapter import DummyAdapter
from dspy import configure, ChatAdapter, LM
import dspy


@pytest.fixture(scope="session")
def event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    try:
        yield loop
    finally:
        loop.close()


# Global variable to track adapter state
_current_adapter_type = None


def get_current_adapter_type():
    """Helper function to check what type of adapter is currently configured"""
    global _current_adapter_type
    return _current_adapter_type


@pytest.fixture(autouse=True, scope="module")
def use_dummy_adapter():
    global _current_adapter_type

    dummy_adapter = DummyAdapter()
    configure(adapter=dummy_adapter)
    _current_adapter_type = "dummy"

    yield

    configure(adapter=ChatAdapter())
    _current_adapter_type = "chat"


if __name__ == "__main__":
    from dspy.adapters import JSONAdapter

    configure(adapter=DummyAdapter())
    print(get_current_adapter_type())

    from elysia.tree.util import DecisionNode
    from elysia.tree.tree import Tree
    from elysia.util.client import ClientManager

    lm = dspy.LM("gpt-4o-mini")

    # class TestSignature(dspy.Signature):
    #     input = dspy.InputField(description="input")
    #     output = dspy.OutputField(description="output")

    # class TestModule(dspy.Module):
    #     def __init__(self):
    #         self.mod = dspy.ChainOfThought(TestSignature)

    #     def forward(self, input: str, lm: dspy.LM):
    #         return self.mod(input=input, lm=lm)

    #     def aforward(self, input: str, lm: dspy.LM):
    #         return self.mod.aforward(input=input, lm=lm)

    # class NonModuleClass:
    #     def __init__(self):
    #         self.mod = TestModule()

    #     def __call__(self, input: str, lm: dspy.LM):
    #         return self.mod(input=input, lm=lm)

    #     async def aforward(self, input: str, lm: dspy.LM):
    #         return await self.mod.aforward(input=input, lm=lm)

    # async def test_non_module_class():
    #     non_module_class = NonModuleClass()
    #     return await non_module_class.aforward(input="hi", lm=lm)

    # print(asyncio.run(test_non_module_class()))

    tree = Tree()

    decision_node = DecisionNode(
        id="test_decision_node",
        instruction="test_instruction",
        options={
            "text_response": {
                "description": "test_description",
                "inputs": {
                    "text": {
                        "description": "response",
                        "type": str,
                        "default": "test_default",
                    }
                },
                "end": True,
                "status": "running",
            }
        },
    )

    async def test_decision_node():

        out = await decision_node(
            tree_data=tree.tree_data,
            base_lm=lm,
            complex_lm=lm,
            available_tools=["text_response"],
            unavailable_tools=[],
            successive_actions={},
            client_manager=ClientManager(),
        )

        return out[0].reasoning

    print(asyncio.run(test_decision_node()))



from typing import Union
from logging import Logger

from rich import print
from rich.panel import Panel

import dspy

from elysia.util.elysia_chain_of_thought import ElysiaChainOfThought
from elysia.objects import Response, Status, Tool, Error
from elysia.tools.retrieval.objects import Aggregation
from elysia.tools.retrieval.prompt_templates import (
    AggregationPrompt,
    construct_aggregation_output_prompt,
)
from elysia.tools.retrieval.util import (
    execute_weaviate_aggregation,
    QueryError,
    AggregationOutput,
    VectorisedAggregationOutput,
)
from elysia.tree.objects import TreeData
from elysia.util.client import ClientManager
from elysia.util.objects import TrainingUpdate, FewShotExamples
from elysia.util.parsing import format_aggregation_response


class Aggregate(Tool):
    def __init__(
        self,
        logger: Logger | None = None,
        **kwargs,
    ):
        self.logger = logger

        super().__init__(
            name="aggregate",
            description="""
            Query the knowledge base specifically for aggregation queries.
            Performs calculations (counting, averaging, summing, etc.) and provides summary statistics on data.
            It can group data by properties and apply filters directly, without needing a prior query.
            Aggregation queries can be filtered.
            This can be applied directly on any collections in the schema.
            Use this tool when you need counts, sums, averages, or other summary statistics on properties in the collections.
            'aggregate' should be considered the first choice for tasks involving counting, summing, averaging, 
            or other statistical operations, even when filtering is required.
            """,
            status="Aggregating...",
            inputs={
                "collection_names": {
                    "description": "the names of the collections that are most relevant to the aggregation. If you are unsure, give all collections.",
                    "type": list[str],
                    "default": [],
                },
            },
            end=False,
        )

    async def is_tool_available(
        self,
        tree_data: TreeData,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager,
    ) -> bool:
        """
        Only available when there is a Weaviate connection.
        If this tool is not available, inform the user that they need to set the WCD_URL and WCD_API_KEY in the settings.
        """
        return client_manager.is_client

    def _find_previous_aggregations(
        self, environment: dict, collection_names: list[str]
    ):
        previous_aggregations = {}
        for collection_name in collection_names:
            if (
                "aggregate" in environment
                and collection_name in environment["aggregate"]
            ):
                for aggregation in environment["aggregate"][collection_name]:
                    if "code" in aggregation and "text" in aggregation["code"]:
                        if collection_name in previous_aggregations:
                            previous_aggregations[collection_name].append(
                                aggregation["code"]["text"]
                            )
                        else:
                            previous_aggregations[collection_name] = [
                                aggregation["code"]["text"]
                            ]

        return previous_aggregations

    def _fix_collection_names(self, collection_names: list[str], schemas: dict):
        return [
            collection_name
            for collection_name in list(schemas.keys())
            if collection_name.lower()
            in [
                schema_collection_name.lower()
                for schema_collection_name in collection_names
            ]
        ]

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        client_manager: ClientManager,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        **kwargs,
    ):

        if self.logger:
            self.logger.debug("Aggregation Tool called!")
            self.logger.debug("Arguments received:")
            self.logger.debug(f"{inputs}")

        # Find matching collection names (to match case sensitivity)
        schemas = tree_data.output_collection_metadata(with_mappings=True)
        if inputs["collection_names"] == [] or inputs["collection_names"] is None:
            collection_names = list(schemas.keys())
        else:
            collection_names = self._fix_collection_names(
                inputs["collection_names"], schemas
            )

        # Set up the dspy model
        aggregation_prompt = AggregationPrompt

        # check if collections are vectorised
        vectorised_by_collection = {
            collection_name: (
                schemas[collection_name]["vectorizer"] is not None
                or (
                    schemas[collection_name]["named_vectors"] is not None
                    and schemas[collection_name]["named_vectors"] != []
                )
            )
            for collection_name in collection_names
        }

        if any(vectorised_by_collection.values()):
            aggregation_prompt = aggregation_prompt.append(
                name="aggregation_queries",
                type_=Union[list[VectorisedAggregationOutput], None],
                field=dspy.OutputField(
                    desc=construct_aggregation_output_prompt(True),
                ),
            )
        else:
            aggregation_prompt = aggregation_prompt.append(
                name="aggregation_queries",
                type_=Union[list[AggregationOutput], None],
                field=dspy.OutputField(
                    desc=construct_aggregation_output_prompt(False),
                ),
            )

        aggregate_generator = ElysiaChainOfThought(
            aggregation_prompt,
            tree_data=tree_data,
            environment=True,
            collection_schemas=True,
            tasks_completed=True,
            message_update=True,
            collection_names=inputs["collection_names"],
            reasoning=tree_data.settings.COMPLEX_USE_REASONING,
        )

        if self.logger:
            self.logger.info(
                f"Any collections vectorised: {any(vectorised_by_collection.values())}"
            )
            self.logger.info(f"Collection names received: {collection_names}")

        # Get some metadata about the collection
        previous_aggregations = self._find_previous_aggregations(
            tree_data.environment.environment, collection_names
        )

        # Generate query with LLM
        try:
            if tree_data.settings.USE_FEEDBACK:
                aggregation, example_uuids = (
                    await aggregate_generator.aforward_with_feedback_examples(
                        feedback_model="aggregate",
                        client_manager=client_manager,
                        base_lm=base_lm,
                        complex_lm=complex_lm,
                        available_collections=collection_names,
                        previous_aggregation_queries=previous_aggregations,
                        num_base_lm_examples=6,
                        return_example_uuids=True,
                    )
                )
            else:
                aggregation = await aggregate_generator.aforward(
                    lm=complex_lm,
                    available_collections=collection_names,
                    previous_aggregation_queries=previous_aggregations,
                )
        except Exception as e:
            yield Error(error_message=str(e))
            return

        if self.logger and aggregation.aggregation_queries is not None:
            self.logger.debug(f"Aggregation: {aggregation.aggregation_queries}")

        # Yield results to front end
        yield Response(text=aggregation.message_update)

        if tree_data.settings.USE_FEEDBACK:
            yield FewShotExamples(uuids=example_uuids)

        # Return if model deems aggregation impossible
        if aggregation.impossible or aggregation.aggregation_queries is None:
            if self.logger:
                self.logger.info(
                    "Model judged this aggregation to be impossible. Returning to the decision tree..."
                )
            for collection_name in collection_names:
                metadata = {
                    "collection_name": collection_name,
                    "impossible": tree_data.user_prompt,
                    "impossible_reasoning": (
                        aggregation.reasoning
                        if tree_data.settings.COMPLEX_USE_REASONING
                        else ""
                    ),
                    "aggregation_output": (
                        aggregation.aggregation_queries.model_dump()
                        if aggregation.aggregation_queries is not None
                        else None
                    ),
                }
                yield Aggregation([], metadata)

            return

        # Go through each response
        for i, aggregation_output in enumerate(aggregation.aggregation_queries):

            aggregation_output.target_collections = self._fix_collection_names(
                aggregation_output.target_collections,
                schemas,
            )

            # Execute query within Weaviate
            async with client_manager.connect_to_async_client() as client:
                try:
                    responses, code_strings = await execute_weaviate_aggregation(
                        client,
                        aggregation_output,
                        property_types={
                            collection_name: {
                                schemas[collection_name]["fields"][i]["name"]: schemas[
                                    collection_name
                                ]["fields"][i]["type"]
                                for i in range(len(schemas[collection_name]["fields"]))
                            }
                            for collection_name in collection_names
                        },
                        vectorised_by_collection=vectorised_by_collection,
                        schema=schemas,
                    )
                except QueryError as e:
                    yield Error(feedback=str(e))
                    continue
                except Exception as e:
                    for collection_name in aggregation_output.target_collections:
                        if self.logger:
                            self.logger.exception(
                                f"Error executing aggregation for {collection_name}: {str(e)}"
                            )
                        yield Error(error_message=str(e))
                    continue  # don't exit, try again for next aggregation output (might not error)

            yield Status(
                f"Aggregated over {', '.join(aggregation_output.target_collections)}"
            )

            for k, (collection_name, response) in enumerate(
                zip(aggregation_output.target_collections, responses)
            ):
                if self.logger and self.logger.level <= 20:
                    print(
                        Panel.fit(
                            code_strings[k],
                            title=f"{collection_name} (Weaviate Aggregation Query)",
                            border_style="yellow",
                            padding=(1, 1),
                        )
                    )

                objects = [
                    {
                        "num_items": (
                            response.total_count
                            if "total_count" in dir(response)
                            else None
                        ),
                        "collections": [
                            {collection_name: format_aggregation_response(response)}
                        ],
                    }
                ]
                metadata = {
                    "collection_name": collection_name,
                    "aggregation_output": aggregation_output.model_dump(),
                    "code": {
                        "language": "python",
                        "title": "Aggregation",
                        "text": code_strings[k],
                    },
                }

                yield Aggregation(objects, metadata)

        # if successful, yield training update
        yield TrainingUpdate(
            module_name="aggregate",
            inputs={
                **tree_data.to_json(),
                "available_collections": collection_names,
                "previous_aggregation_queries": previous_aggregations,
            },
            outputs=aggregation.__dict__["_store"],
        )
        if self.logger:
            self.logger.debug("Aggregation Tool finished!")



import os
import pytest
from elysia.config import Settings, configure

from elysia.config import settings as global_settings
from elysia.config import reset_settings

from elysia.tree import Tree


def test_smart_setup():
    """
    Test that the smart setup is correct
    """
    settings = Settings.from_smart_setup()
    assert settings.BASE_MODEL is not None
    assert settings.COMPLEX_MODEL is not None
    assert settings.BASE_PROVIDER is not None
    assert settings.COMPLEX_PROVIDER is not None



import psutil

from elysia.api.services.user import UserManager


def get_total_memory():
    return round(psutil.virtual_memory().total / (1024 * 1024), 2)


def get_free_memory():
    return round(psutil.virtual_memory().available / (1024 * 1024), 2)


def get_cpu_usage():
    return psutil.cpu_percent(interval=1)


def get_memory_usage():
    return psutil.virtual_memory().percent


def get_disk_usage():
    return psutil.disk_usage("/").percent


def get_number_local_users(user_manager: UserManager):
    return len(user_manager.users)


async def get_average_user_memory(user_manager: UserManager):
    if len(user_manager.users) == 0:
        return 0, 0
    avg_user_memory = 0
    avg_tree_memory = 0
    for user in user_manager.users.values():
        user_memory = 0
        for tree in user["tree_manager"].trees.values():
            user_memory += tree["tree"].detailed_memory_usage()["total"] / (1024 * 1024)

        if len(user["tree_manager"].trees) > 0:
            avg_tree_memory += user_memory / len(user["tree_manager"].trees)

        avg_user_memory += user_memory

    return (
        round(avg_user_memory / len(user_manager.users), 2),
        round(avg_tree_memory / len(user_manager.users), 2),
    )


# async def get_average_user_requests(user_manager: UserManager):
#     return await user_manager.get_average_user_requests()


async def print_resources(
    user_manager: UserManager | None = None, save_to_file: bool = False
):
    if user_manager is not None:
        avg_user_memory, avg_tree_memory = await get_average_user_memory(user_manager)
        # avg_user_requests = await get_average_user_requests(user_manager)
        # num_users_db = await get_number_local_users_db(user_manager)

    print("\n\n\n\n")
    print("-" * 100)
    print(f"ELYSIA RESOURCES")
    print("-" * 100)
    print(f"Total memory: {get_total_memory()} MB")
    print(f"Free memory: {get_free_memory()} MB")
    print(f"CPU usage: {get_cpu_usage()}%")
    print(f"Memory usage: {get_memory_usage()}%")
    print(f"Disk usage: {get_disk_usage()}%")
    if user_manager is not None:
        print("\n\n")
        print("-" * 100)
        print(f"USER STATISTICS")
        print("-" * 100)
        print(f"Number of local users: {get_number_local_users(user_manager)}")
        # print(f"Number of unique users in database: {num_users_db}")
        print(f"Average user memory usage: {avg_user_memory} MB")
        print(f"Average tree memory usage: {avg_tree_memory} MB")
        # print(f"Average user requests: {avg_user_requests}")
    print("-" * 100)
    print("\n\n\n\n")

    if save_to_file:
        with open("resources.txt", "w") as f:
            f.write("-" * 100)
            f.write("\n")
            f.write(f"ELYSIA RESOURCES\n")
            f.write("-" * 100)
            f.write("\n")
            f.write(f"Total memory: {get_total_memory()} MB\n")
            f.write(f"Free memory: {get_free_memory()} MB\n")
            f.write(f"CPU usage: {get_cpu_usage()}%\n")
            f.write(f"Memory usage: {get_memory_usage()}%\n")
            f.write(f"Disk usage: {get_disk_usage()}%\n")
            if user_manager is not None:
                f.write("\n\n")
                f.write("-" * 100)
                f.write("\nUSER STATISTICS\n")
                f.write("-" * 100)
                f.write("\n\n")
                f.write(
                    f"Number of local users: {get_number_local_users(user_manager)}\n"
                )
                # f.write(f"Number of unique users in database: {num_users_db}\n")
                f.write(f"Average user memory usage: {avg_user_memory} MB\n")
                f.write(f"Average tree memory usage: {avg_tree_memory} MB\n")
                # f.write(f"Average user requests: {avg_user_requests}\n")


# if __name__ == "__main__":
#     await print_resources()



import pytest
import os

from weaviate.util import generate_uuid5
from elysia.api.services.user import UserManager
from elysia.api.api_types import (
    InitialiseTreeData,
    AddToolToTreeData,
    RemoveToolFromTreeData,
    AddBranchToTreeData,
    RemoveBranchFromTreeData,
    SaveConfigUserData,
    UpdateFrontendConfigData,
)
from elysia.api.routes.query import process
from elysia.api.dependencies.common import get_user_manager
from elysia.api.routes.init import initialise_user, initialise_tree
from elysia.api.routes.user_config import (
    update_frontend_config,
    save_config_user,
    get_current_user_config,
)
from elysia.api.routes.tools import (
    get_available_tools,
    add_tool_to_tree,
    remove_tool_from_tree,
    add_branch_to_tree,
    remove_branch_from_tree,
)
from elysia.api.custom_tools import TellAJoke
from elysia.api.dependencies.common import get_user_manager
from elysia.api.routes.init import initialise_user, initialise_tree
from elysia.api.routes.user_config import (
    update_frontend_config,
    save_config_user,
    get_current_user_config,
)
from elysia.api.routes.query import process

from fastapi.responses import JSONResponse
import json


def read_response(response: JSONResponse):
    return json.loads(response.body)


from elysia.api.core.log import logger, set_log_level

set_log_level("CRITICAL")


def read_response(response: JSONResponse):
    return json.loads(response.body)


async def initialise_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()

    response = await initialise_user(
        user_id,
        user_manager,
    )

    response = await initialise_tree(
        user_id,
        conversation_id,
        InitialiseTreeData(
            low_memory=False,
        ),
        user_manager,
    )


@pytest.mark.asyncio
async def test_tools_exist():
    joke_tool = TellAJoke()
    response = await get_available_tools()
    response = read_response(response)
    assert response["error"] == ""
    assert "tools" in response
    assert response["tools"] is not None
    assert "TellAJoke" in response["tools"]
    assert response["tools"]["TellAJoke"]["description"] == joke_tool.description
    assert response["tools"]["TellAJoke"]["inputs"].keys() == joke_tool.inputs.keys()
    assert response["tools"]["TellAJoke"]["end"] == joke_tool.end
    assert response["tools"]["TellAJoke"]["name"] == joke_tool.name


@pytest.mark.asyncio
async def test_basic_add_tool_to_tree():
    user_id = "test_user_basic_add_tool_to_tree"
    conversation_id = "test_conversation_basic_add_tool_to_tree"

    user_manager = get_user_manager()
    await initialise_user_and_tree(user_id, conversation_id)

    tool = TellAJoke()

    # add tool to tree
    response = await add_tool_to_tree(
        user_id=user_id,
        data=AddToolToTreeData(
            tool_name="TellAJoke",
            branch_id="base",
            from_tool_ids=[],
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # check returned tree is correct
    assert tool.name in response["tree"]["options"]
    assert response["tree"]["options"][tool.name]["description"] == tool.description
    assert response["tree"]["options"][tool.name]["id"] == tool.name
    assert response["tree"]["options"][tool.name]["branch"] is False

    # check tree itself is correct
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tool.name in tree.tree["options"]
    assert tree.tree["options"][tool.name]["description"] == tool.description
    assert tree.tree["options"][tool.name]["id"] == tool.name
    assert tree.tree["options"][tool.name]["branch"] is False


@pytest.mark.asyncio
async def test_add_tool_to_tree_with_from_tool_ids():
    user_id = "test_user_add_tool_to_tree_with_from_tool_ids"
    conversation_id = "test_conversation_add_tool_to_tree_with_from_tool_ids"

    user_manager = get_user_manager()

    tool = TellAJoke()

    await initialise_user_and_tree(user_id, conversation_id)

    # add tool to tree
    response = await add_tool_to_tree(
        user_id=user_id,
        data=AddToolToTreeData(
            tool_name="TellAJoke",
            branch_id="base",
            from_tool_ids=["query"],
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # check returned tree is correct
    assert tool.name in response["tree"]["options"]["query"]["options"]
    assert (
        response["tree"]["options"]["query"]["options"][tool.name]["description"]
        == tool.description
    )
    assert response["tree"]["options"]["query"]["options"][tool.name]["id"] == tool.name
    assert response["tree"]["options"]["query"]["options"][tool.name]["branch"] is False

    # check tree itself is correct
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tool.name in tree.tree["options"]["query"]["options"]
    assert (
        tree.tree["options"]["query"]["options"][tool.name]["description"]
        == tool.description
    )
    assert tree.tree["options"]["query"]["options"][tool.name]["id"] == tool.name
    assert tree.tree["options"]["query"]["options"][tool.name]["branch"] is False


@pytest.mark.asyncio
async def test_basic_remove_tool_from_tree():
    user_id = "test_user_basic_remove_tool_from_tree"
    conversation_id = "test_conversation_basic_remove_tool_from_tree"

    user_manager = get_user_manager()

    tool = TellAJoke()

    await initialise_user_and_tree(user_id, conversation_id)

    # first add tool to tree
    response = await add_tool_to_tree(
        user_id=user_id,
        data=AddToolToTreeData(
            tool_name="TellAJoke",
            branch_id="base",
            from_tool_ids=[],
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # remove tool from tree
    response = await remove_tool_from_tree(
        user_id=user_id,
        data=RemoveToolFromTreeData(
            tool_name="TellAJoke",
            branch_id="base",
            from_tool_ids=[],
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # check returned tree is correct
    assert tool.name not in response["tree"]["options"]

    # check tree itself is correct
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tool.name not in tree.tree["options"]


@pytest.mark.asyncio
async def test_remove_tool_from_tree_with_from_tool_ids():
    user_id = "test_user_remove_tool_from_tree_with_from_tool_ids"
    conversation_id = "test_conversation_remove_tool_from_tree_with_from_tool_ids"

    user_manager = get_user_manager()

    tool = TellAJoke()

    await initialise_user_and_tree(user_id, conversation_id)

    # first add tool to tree
    response = await add_tool_to_tree(
        user_id=user_id,
        data=AddToolToTreeData(
            tool_name="TellAJoke",
            branch_id="base",
            from_tool_ids=["query"],
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # remove tool from tree
    response = await remove_tool_from_tree(
        user_id=user_id,
        data=RemoveToolFromTreeData(
            tool_name="TellAJoke",
            branch_id="base",
            from_tool_ids=["query"],
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # check returned tree is correct
    assert tool.name not in response["tree"]["options"]["query"]["options"]

    # check tree itself is correct
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tool.name not in tree.tree["options"]["query"]["options"]


@pytest.mark.asyncio
async def test_basic_add_branch_to_tree():
    user_id = "test_user_basic_add_branch_to_tree"
    conversation_id = "test_conversation_basic_add_branch_to_tree"

    user_manager = get_user_manager()

    await initialise_user_and_tree(user_id, conversation_id)

    # add branch to tree
    response = await add_branch_to_tree(
        user_id=user_id,
        data=AddBranchToTreeData(
            id="new_branch",
            description="New Branch Description",
            instruction="New Branch Instruction",
            from_branch_id="base",
            from_tool_ids=[],
            status="New Branch Status",
            root=False,
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # check returned tree is correct
    assert "new_branch" in response["tree"]["options"]
    assert response["tree"]["options"]["new_branch"]["id"] == "new_branch"
    assert (
        response["tree"]["options"]["new_branch"]["description"]
        == "New Branch Description"
    )
    assert (
        response["tree"]["options"]["new_branch"]["instruction"]
        == "New Branch Instruction"
    )
    assert response["tree"]["options"]["new_branch"]["branch"] is True
    assert response["tree"]["options"]["new_branch"]["options"] == {}

    # check tree itself is correct
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert "new_branch" in tree.tree["options"]
    assert tree.tree["options"]["new_branch"]["id"] == "new_branch"
    assert tree.tree["options"]["new_branch"]["description"] == "New Branch Description"
    assert tree.tree["options"]["new_branch"]["instruction"] == "New Branch Instruction"
    assert tree.tree["options"]["new_branch"]["branch"] is True
    assert tree.tree["options"]["new_branch"]["options"] == {}


@pytest.mark.asyncio
async def test_add_branch_to_branch():
    user_id = "test_user_add_branch_to_branch"
    conversation_id = "test_conversation_add_branch_to_branch"
    user_manager = get_user_manager()

    await initialise_user_and_tree(user_id, conversation_id)

    # get config_id
    response = await get_current_user_config(
        user_id,
        user_manager,
    )
    assert read_response(response)["error"] == ""
    config_id = read_response(response)["config"]["id"]
    config_name = read_response(response)["config"]["name"]

    # change branch initialisation
    response = await save_config_user(
        user_id,
        config_id,
        SaveConfigUserData(
            name=config_name,
            config={
                "branch_initialisation": "multi_branch",
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
            default=True,
        ),
        user_manager,
    )
    assert read_response(response)["error"] == ""

    # add branch to tree
    response = await add_branch_to_tree(
        user_id=user_id,
        data=AddBranchToTreeData(
            id="new_branch",
            description="New Branch Description",
            instruction="New Branch Instruction",
            from_branch_id="search",
            status="New Branch Status",
            from_tool_ids=[],
            root=False,
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    assert response["tree"] is not None
    assert response["tree"] is not {}

    # check returned tree is correct
    assert "new_branch" in response["tree"]["options"]["search"]["options"]
    assert (
        response["tree"]["options"]["search"]["options"]["new_branch"]["id"]
        == "new_branch"
    )
    assert (
        response["tree"]["options"]["search"]["options"]["new_branch"]["description"]
        == "New Branch Description"
    )
    assert (
        response["tree"]["options"]["search"]["options"]["new_branch"]["instruction"]
        == "New Branch Instruction"
    )
    assert (
        response["tree"]["options"]["search"]["options"]["new_branch"]["branch"] is True
    )
    assert (
        response["tree"]["options"]["search"]["options"]["new_branch"]["options"] == {}
    )

    # check tree itself is correct
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert "new_branch" in tree.tree["options"]["search"]["options"]
    assert tree.tree["options"]["search"]["options"]["new_branch"]["id"] == "new_branch"
    assert (
        tree.tree["options"]["search"]["options"]["new_branch"]["description"]
        == "New Branch Description"
    )
    assert (
        tree.tree["options"]["search"]["options"]["new_branch"]["instruction"]
        == "New Branch Instruction"
    )
    assert tree.tree["options"]["search"]["options"]["new_branch"]["branch"] is True
    assert tree.tree["options"]["search"]["options"]["new_branch"]["options"] == {}


@pytest.mark.asyncio
async def test_full_cycle():
    user_id = "test_user_full_cycle"
    conversation_id = "test_conversation_full_cycle"

    user_manager = get_user_manager()

    await initialise_user_and_tree(user_id, conversation_id)

    # get config_id
    response = await get_current_user_config(
        user_id,
        user_manager,
    )
    assert read_response(response)["error"] == ""
    config_id = read_response(response)["config"]["id"]
    config_name = read_response(response)["config"]["name"]

    # change branch initialisation
    response = await save_config_user(
        user_id,
        config_id,
        SaveConfigUserData(
            name=config_name,
            config={
                "branch_initialisation": "multi_branch",
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
            default=True,
        ),
        user_manager,
    )
    assert read_response(response)["error"] == ""

    # add a branch to a tool
    response = await add_branch_to_tree(
        user_id=user_id,
        data=AddBranchToTreeData(
            id="new_branch",
            description="New Branch Description",
            instruction="New Branch Instruction",
            from_branch_id="search",
            from_tool_ids=["query"],
            status="New Branch Status",
            root=False,
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # check the tree is correct
    assert response["tree"] is not None
    assert response["tree"] != {}
    assert (
        "new_branch"
        in response["tree"]["options"]["search"]["options"]["query"]["options"]
    )
    assert (
        response["tree"]["options"]["search"]["options"]["query"]["options"][
            "new_branch"
        ]["id"]
        == "new_branch"
    )

    # get tree and check there is the new branch
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert "new_branch" in tree.tree["options"]["search"]["options"]["query"]["options"]
    assert (
        tree.tree["options"]["search"]["options"]["query"]["options"]["new_branch"][
            "id"
        ]
        == "new_branch"
    )

    # check decision nodes updated
    assert "new_branch" in tree.decision_nodes
    assert "search.query" in tree.decision_nodes
    assert "new_branch" in tree.decision_nodes["search.query"].options

    # add a tool to the new branch
    response = await add_tool_to_tree(
        user_id=user_id,
        data=AddToolToTreeData(
            tool_name="TellAJoke",
            branch_id="new_branch",
            from_tool_ids=[],
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # check the tree is correct
    assert (
        "tell_a_joke"
        in response["tree"]["options"]["search"]["options"]["query"]["options"][
            "new_branch"
        ]["options"]
    )

    # and tree itself is correct
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert (
        "tell_a_joke"
        in tree.tree["options"]["search"]["options"]["query"]["options"]["new_branch"][
            "options"
        ]
    )

    # check decision nodes updated
    assert "tell_a_joke" in tree.decision_nodes["new_branch"].options

    # remove the branch from the tree
    response = await remove_branch_from_tree(
        user_id=user_id,
        data=RemoveBranchFromTreeData(id="new_branch"),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # check the branch and its stemmed tools are gone
    assert (
        "new_branch"
        not in response["tree"]["options"]["search"]["options"]["query"]["options"]
    )

    tree = await user_manager.get_tree(user_id, conversation_id)
    assert (
        "new_branch"
        not in tree.tree["options"]["search"]["options"]["query"]["options"]
    )
    assert "new_branch" not in tree.decision_nodes



import os
import pytest

from elysia.config import Settings
from elysia.api.routes.init import initialise_user
from elysia.api.routes.user_config import (
    load_a_config,
    save_config_user,
    load_config_user,
    new_user_config,
    get_current_user_config,
)
from elysia.api.api_types import SaveConfigUserData
from fastapi.responses import JSONResponse
import json
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from weaviate.util import generate_uuid5


def read_response(response: JSONResponse):
    return json.loads(response.body)


async def delete_config_after_completion(
    user_manager: UserManager, user_id: str, config_id: str
):
    client_manager = user_manager.users[user_id]["client_manager"]
    async with client_manager.connect_to_async_client() as client:
        if not await client.collections.exists("ELYSIA_CONFIG__"):
            return

        collection = client.collections.get("ELYSIA_CONFIG__")
        uuid = generate_uuid5(config_id)
        if await collection.data.exists(uuid):
            await collection.data.delete_by_id(uuid)

    if os.path.exists(f"elysia/api/user_configs/frontend_config_{user_id}.json"):
        os.remove(f"elysia/api/user_configs/frontend_config_{user_id}.json")


@pytest.mark.asyncio
async def test_encryption():

    user_id = "test_encryption"

    user_manager = get_user_manager()

    # initialise the user
    response = await initialise_user(user_id=user_id, user_manager=user_manager)
    response = read_response(response)

    # create a new config
    response = await new_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    config_id = response["config"]["id"]

    # check the WCD_API_KEY
    assert response["config"]["settings"]["WCD_API_KEY"] == os.getenv("WCD_API_KEY")

    # change config, don't change anything
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=response["config"]["name"],
            default=True,
            config=response["config"],
            frontend_config={},
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["config"]["settings"]["WCD_API_KEY"] == os.getenv("WCD_API_KEY")

    # get the config
    response = await load_a_config(
        user_id=user_id, config_id=config_id, user_manager=user_manager
    )
    response = read_response(response)
    assert response["config"]["settings"]["WCD_API_KEY"] == os.getenv("WCD_API_KEY")


@pytest.mark.asyncio
async def test_incorrect_api_keys():

    user_id = "test_incorrect_api_keys"
    user_manager = get_user_manager()

    correct_api_key = os.getenv("OPENROUTER_API_KEY")
    incorrect_api_key = "incorrect"

    # initialise the user
    response = await initialise_user(user_id=user_id, user_manager=user_manager)
    response = read_response(response)

    # create a new config
    response = await new_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    config_id = response["config"]["id"]

    # change config, change the OPENROUTER_API_KEY
    new_config = response["config"]
    # new_config["settings"]["OPENROUTER_API_KEY"] = incorrect_api_key
    new_config["settings"]["API_KEYS"]["openrouter_api_key"] = incorrect_api_key

    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=response["config"]["name"],
            default=True,
            config=new_config,
            frontend_config={},
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert (
        response["config"]["settings"]["API_KEYS"]["openrouter_api_key"]
        == incorrect_api_key
    )

    # load a config
    response = await get_current_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    assert (
        response["config"]["settings"]["API_KEYS"]["openrouter_api_key"]
        == incorrect_api_key
    )


@pytest.mark.asyncio
async def test_change_models():

    user_id = "test_change_models"
    user_manager = get_user_manager()

    # initialise the user
    response = await initialise_user(user_id=user_id, user_manager=user_manager)
    response = read_response(response)

    # create a new config
    response = await new_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    config_id = response["config"]["id"]

    # change config, change the models
    new_config = response["config"]
    new_config["settings"]["BASE_MODEL"] = "gpt-4o"
    new_config["settings"]["COMPLEX_MODEL"] = "gpt-4o"
    new_config["settings"]["BASE_PROVIDER"] = "openrouter/openai"
    new_config["settings"]["COMPLEX_PROVIDER"] = "openrouter/openai"

    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=response["config"]["name"],
            default=True,
            config=new_config,
            frontend_config={},
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["config"]["settings"]["BASE_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["COMPLEX_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["BASE_PROVIDER"] == "openrouter/openai"
    assert response["config"]["settings"]["COMPLEX_PROVIDER"] == "openrouter/openai"

    # load a config
    response = await get_current_user_config(user_id=user_id, user_manager=user_manager)
    response = read_response(response)
    assert response["config"]["settings"]["BASE_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["COMPLEX_MODEL"] == "gpt-4o"
    assert response["config"]["settings"]["BASE_PROVIDER"] == "openrouter/openai"
    assert response["config"]["settings"]["COMPLEX_PROVIDER"] == "openrouter/openai"



models = {
    "openai": {
        "gpt-4.1": {
            "name": "gpt-4.1",
            "api_keys": ["openai_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "gpt-4.1-mini": {
            "name": "gpt-4.1-mini",
            "api_keys": ["openai_api_key"],
            "speed": "medium",
            "accuracy": "medium",
        },
        "gpt-4.1-nano": {
            "name": "gpt-4.1-nano",
            "api_keys": ["openai_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "anthropic": {
        "claude-sonnet-4-20250514": {
            "name": "claude-sonnet-4-20250514",
            "api_keys": ["anthropic_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-7-sonnet-20250219": {
            "name": "claude-3-7-sonnet-20250219",
            "api_keys": ["anthropic_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-5-haiku-20241022": {
            "name": "claude-3-5-haiku-20241022",
            "api_keys": ["anthropic_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "openrouter/openai": {
        "gpt-4.1": {
            "name": "gpt-4.1",
            "api_keys": ["openai_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "gpt-4.1-mini": {
            "name": "gpt-4.1-mini",
            "api_keys": ["openai_api_key"],
            "speed": "medium",
            "accuracy": "medium",
        },
        "gpt-4.1-nano": {
            "name": "gpt-4.1-nano",
            "api_keys": ["openai_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "openrouter/anthropic": {
        "claude-sonnet-4": {
            "name": "claude-sonnet-4",
            "api_keys": ["openrouter_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-7-sonnet": {
            "name": "claude-3-7-sonnet",
            "api_keys": ["openrouter_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
        "claude-3-5-haiku": {
            "name": "claude-3-5-haiku",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
    },
    "openrouter/google": {
        "gemini-2.5-flash": {
            "name": "gemini-2.5-flash",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-flash-lite": {
            "name": "gemini-2.5-flash-lite",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
        "gemini-2.0-flash-001": {
            "name": "gemini-2.0-flash-001",
            "api_keys": ["openrouter_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-pro": {
            "name": "gemini-2.5-pro",
            "api_keys": ["openrouter_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
    },
    "gemini": {
        "gemini-2.5-flash": {
            "name": "gemini-2.5-flash",
            "api_keys": ["gemini_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-flash-lite": {
            "name": "gemini-2.5-flash-lite",
            "api_keys": ["gemini_api_key"],
            "speed": "fast",
            "accuracy": "low",
        },
        "gemini-2.0-flash-001": {
            "name": "gemini-2.0-flash-001",
            "api_keys": ["gemini_api_key"],
            "speed": "fast",
            "accuracy": "medium",
        },
        "gemini-2.5-pro": {
            "name": "gemini-2.5-pro",
            "api_keys": ["gemini_api_key"],
            "speed": "slow",
            "accuracy": "high",
        },
    },
}
